"""
:mod:`neo.rawio` provides classes for reading
electrophysiological data files with a low-level API

:attr:`neo.rawio.rawiolist` provides a list of successfully imported rawio
classes.

Functions:

.. autofunction:: neo.rawio.get_rawio


Classes:

* :attr:`AlphaOmegaRawIO`
* :attr:`AxographRawIO`
* :attr:`AxonaRawIO`
* :attr:`AxonRawIO`
* :attr:`BiocamRawIO`
* :attr:`BlackrockRawIO`
* :attr:`BrainVisionRawIO`
* :attr:`CedRawIO`
* :attr:`EDFRawIO`
* :attr:`ElanRawIO`
* :attr:`IntanRawIO`
* :attr:`MaxwellRawIO`
* :attr:`MedRawIO`
* :attr:`MEArecRawIO`
* :attr:`MicromedRawIO`
* :attr:`NeuralynxRawIO`
* :attr:`NeuroExplorerRawIO`
* :attr:`NeuroScopeRawIO`
* :attr:`NIXRawIO`
* :attr:`OpenEphysRawIO`
* :attr:`OpenEphysBinaryRawIO`
* :attr:`PhyRawIO`
* :attr:`PlexonRawIO`
* :attr:`Plexon2RawIO`
* :attr:`RawBinarySignalRawIO`
* :attr:`RawMCSRawIO`
* :attr:`Spike2RawIO`
* :attr:`SpikeGadgetsRawIO`
* :attr:`SpikeGLXRawIO`
* :attr:`TdtRawIO`
* :attr:`WinEdrRawIO`
* :attr:`WinWcpRawIO`


.. autoclass:: neo.rawio.AlphaOmegaRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.AxographRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.AxonaRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.AxonRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.BiocamRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.BlackrockRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.BrainVisionRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.CedRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.EDFRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.ElanRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.IntanRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.MaxwellRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.MedRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.MEArecRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.MicromedRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.NeuralynxRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.NeuroExplorerRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.NeuroScopeRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.NIXRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.OpenEphysRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.OpenEphysBinaryRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.PhyRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.PlexonRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.Plexon2RawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.RawBinarySignalRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.RawMCSRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.Spike2RawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.SpikeGadgetsRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.SpikeGLXRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.TdtRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.WinEdrRawIO

    .. autoattribute:: extensions

.. autoclass:: neo.rawio.WinWcpRawIO

    .. autoattribute:: extensions

"""

from pathlib import Path
from collections import Counter

from neo.rawio.alphaomegarawio import AlphaOmegaRawIO
from neo.rawio.axographrawio import AxographRawIO
from neo.rawio.axonarawio import AxonaRawIO
from neo.rawio.axonrawio import AxonRawIO
from neo.rawio.biocamrawio import BiocamRawIO
from neo.rawio.blackrockrawio import BlackrockRawIO
from neo.rawio.brainvisionrawio import BrainVisionRawIO
from neo.rawio.cedrawio import CedRawIO
from neo.rawio.edfrawio import EDFRawIO
from neo.rawio.elanrawio import ElanRawIO
from neo.rawio.examplerawio import ExampleRawIO
from neo.rawio.intanrawio import IntanRawIO
from neo.rawio.maxwellrawio import MaxwellRawIO
from neo.rawio.mearecrawio import MEArecRawIO
from neo.rawio.medrawio import MedRawIO
from neo.rawio.micromedrawio import MicromedRawIO
from neo.rawio.neuralynxrawio import NeuralynxRawIO
from neo.rawio.neuroexplorerrawio import NeuroExplorerRawIO
from neo.rawio.neuroscoperawio import NeuroScopeRawIO
from neo.rawio.nixrawio import NIXRawIO
from neo.rawio.openephysrawio import OpenEphysRawIO
from neo.rawio.openephysbinaryrawio import OpenEphysBinaryRawIO
from neo.rawio.phyrawio import PhyRawIO
from neo.rawio.plexonrawio import PlexonRawIO
from neo.rawio.plexon2rawio import Plexon2RawIO
from neo.rawio.rawbinarysignalrawio import RawBinarySignalRawIO
from neo.rawio.rawmcsrawio import RawMCSRawIO
from neo.rawio.spike2rawio import Spike2RawIO
from neo.rawio.spikegadgetsrawio import SpikeGadgetsRawIO
from neo.rawio.spikeglxrawio import SpikeGLXRawIO
from neo.rawio.tdtrawio import TdtRawIO
from neo.rawio.winedrrawio import WinEdrRawIO
from neo.rawio.winwcprawio import WinWcpRawIO

rawiolist = [
    AlphaOmegaRawIO,
    AxographRawIO,
    AxonaRawIO,
    AxonRawIO,
    BiocamRawIO,
    BlackrockRawIO,
    BrainVisionRawIO,
    CedRawIO,
    EDFRawIO,
    ElanRawIO,
    IntanRawIO,
    MicromedRawIO,
    MaxwellRawIO,
    MEArecRawIO,
    MedRawIO,
    NeuralynxRawIO,
    NeuroExplorerRawIO,
    NeuroScopeRawIO,
    NIXRawIO,
    OpenEphysRawIO,
    OpenEphysBinaryRawIO,
    PhyRawIO,
    PlexonRawIO,
    Plexon2RawIO,
    RawBinarySignalRawIO,
    RawMCSRawIO,
    Spike2RawIO,
    SpikeGadgetsRawIO,
    SpikeGLXRawIO,
    TdtRawIO,
    WinEdrRawIO,
    WinWcpRawIO,
]


def get_rawio_class(filename_or_dirname):
    """Legacy function for returning class guess from file extension
    DEPRECATED
    """

    import warnings

    warnings.warn(
        "get_rawio_class is deprecated and will be removed in 0.15.0. " "In the future please use `get_rawio`"
    )

    return get_rawio(filename_or_dirname)


def get_rawio(filename_or_dirname, exclusive_rawio: bool = True):
    """
    Return a neo.rawio class guess from file extension.

    Parameters
    ----------
    filename_or_dirname : str | Path
        The filename or directory name to check for file suffixes that
        can be read by Neo. This can also be used to check whether a
        rawio could read a not-yet written file
    exclusive_rawio: bool, default: True
        Whether to return a rawio if there is only one rawio capable of
        reading the file. If this doesn't exist will return None.
        If set to False it will return all possible rawios organized
        by the most likely rawio.

    Returns
    -------
    possibles: neo.RawIO | None | list[neo.RawIO]
        If exclusive_rawio is True, returns the single RawIO that
        can read a file/set of files or None. If exclusive_rawio is
        False it will return all possible RawIOs (organized by most likely)
        that could read the file or files.
    """
    filename_or_dirname = Path(filename_or_dirname)

    # if filename_or_dirname doesn't exist then user may just be checking if
    # neo can read their file or they give a real file
    if not filename_or_dirname.exists() or filename_or_dirname.is_file():
        ext = Path(filename_or_dirname).suffix
        ext_list = [ext[1:]]
    else:
        ext_list = list({filename.suffix[1:] for filename in filename_or_dirname.glob("*") if filename.is_file()})

    possibles = []
    for ext in ext_list:
        for rawio in rawiolist:
            if any(ext.lower() == ext2.lower() for ext2 in rawio.extensions):
                possibles.append(rawio)

    if len(possibles) == 1 and exclusive_rawio:
        return possibles[0]
    elif exclusive_rawio:
        return None
    else:
        possibles = [io[0] for io in Counter(possibles).most_common()]
        return possibles



"""
Class for reading data from a .kwik dataset

Depends on: scipy
            phy

Supported: Read

Author: Mikkel E. Lepperød @CINPLA

"""

# TODO: writing to file

import numpy as np
import quantities as pq
import os

# I need to subclass BaseIO
from neo.io.baseio import BaseIO

# to import from core
from neo.core import Segment, SpikeTrain, AnalogSignal, Block, Group


class KwikIO(BaseIO):
    """
    Class for "reading" experimental data from a .kwik file.

    Generates a :class:`Segment` with a :class:`AnalogSignal`

    """

    is_readable = True  # This class can only read data
    is_writable = False  # write is not supported

    supported_objects = [Block, Segment, SpikeTrain, AnalogSignal, Group]

    # This class can return either a Block or a Segment
    # The first one is the default ( self.read )
    # These lists should go from highest object to lowest object because
    # common_io_test assumes it.
    readable_objects = [Block]

    # This class is not able to write objects
    writeable_objects = []

    has_header = False
    is_streameable = False

    name = "Kwik"
    description = "This IO reads experimental data from a .kwik dataset"
    extensions = ["kwik"]
    mode = "file"

    def __init__(self, filename):
        """
        Arguments:
            filename : the filename
        """
        from klusta import kwik

        BaseIO.__init__(self)
        self.filename = os.path.abspath(filename)
        model = kwik.KwikModel(self.filename)  # TODO this group is loaded twice
        self.models = [kwik.KwikModel(self.filename, channel_group=grp) for grp in model.channel_groups]

    def read_block(
        self,
        lazy=False,
        get_waveforms=True,
        cluster_group=None,
        raw_data_units="uV",
        get_raw_data=False,
    ):
        """
        Reads a block with segments and groups

        Parameters
        ----------
        get_waveforms: bool, default = False
            Wether or not to get the waveforms
        get_raw_data: bool, default = False
            Wether or not to get the raw traces
        raw_data_units: str, default = "uV"
            SI units of the raw trace according to voltage_gain given to klusta
        cluster_group: str, default = None
            Which clusters to load, possibilities are "noise", "unsorted",
            "good", if None all is loaded.
        """
        assert not lazy, "Do not support lazy"

        blk = Block()
        seg = Segment(file_origin=self.filename)
        blk.segments += [seg]
        for model in self.models:
            group_id = model.channel_group
            group_meta = {"group_id": group_id}
            group_meta.update(model.metadata)
            chx = Group(name=f"channel group #{group_id}", index=model.channels, **group_meta)
            blk.groups.append(chx)
            clusters = model.spike_clusters
            for cluster_id in model.cluster_ids:
                meta = model.cluster_metadata[cluster_id]
                if cluster_group is None:
                    pass
                elif cluster_group != meta:
                    continue
                sptr = self.read_spiketrain(
                    cluster_id=cluster_id, model=model, get_waveforms=get_waveforms, raw_data_units=raw_data_units
                )
                sptr.annotations.update({"cluster_group": meta, "group_id": model.channel_group})
                unit = Group(cluster_group=meta, group_id=model.channel_group, name=f"unit #{cluster_id}")
                unit.add(sptr)
                chx.add(unit)
                seg.spiketrains.append(sptr)
            if get_raw_data:
                ana = self.read_analogsignal(model, units=raw_data_units)
                chx.add(ana)
                seg.analogsignals.append(ana)

        seg.duration = model.duration * pq.s

        blk.check_relationships()
        return blk

    def read_analogsignal(self, model, units="uV", lazy=False):
        """
        Reads analogsignals

        Parameters
        ----------
        units: str, default = "uV"
            SI units of the raw trace according to voltage_gain given to klusta
        """
        assert not lazy, "Do not support lazy"

        arr = model.traces[:] * model.metadata["voltage_gain"]
        ana = AnalogSignal(
            arr, sampling_rate=model.sample_rate * pq.Hz, units=units, file_origin=model.metadata["raw_data_files"]
        )
        return ana

    def read_spiketrain(self, cluster_id, model, lazy=False, get_waveforms=True, raw_data_units=None):
        """
        Reads sorted spiketrains

        Parameters
        ----------
        get_waveforms: bool, default = False
            Wether or not to get the waveforms
        cluster_id: int,
            Which cluster to load, according to cluster id from klusta
        model: klusta.kwik.KwikModel
            A KwikModel object obtained by klusta.kwik.KwikModel(fname)
        """
        try:
            if not (cluster_id in model.cluster_ids):
                raise ValueError
        except ValueError:
            print(f"Exception: cluster_id ({cluster_id}) not found !! ")
            return
        clusters = model.spike_clusters
        idx = np.nonzero(clusters == cluster_id)
        if get_waveforms:
            w = model.all_waveforms[idx]
            # klusta: num_spikes, samples_per_spike, num_chans = w.shape
            w = w.swapaxes(1, 2)
            w = pq.Quantity(w, raw_data_units)
        else:
            w = None
        if model.duration > 0.0:
            t_stop = model.duration
        else:
            t_stop = np.max(model.spike_times[idx])
        sptr = SpikeTrain(
            times=model.spike_times[idx],
            t_stop=t_stop,
            waveforms=w,
            units="s",
            sampling_rate=model.sample_rate * pq.Hz,
            file_origin=self.filename,
            **{"cluster_id": cluster_id},
        )
        return sptr



"""
AxographIO
==========

IO class for reading AxoGraph files (.axgd, .axgx)
"""

from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.axographrawio import AxographRawIO


class AxographIO(AxographRawIO, BaseFromRaw):
    """
    IO class for reading AxoGraph files (.axgd, .axgx)

    Args:
        filename (string):
            File name of the AxoGraph file to read.
        force_single_segment (bool):
            Episodic files are normally read as multi-Segment Neo objects. This
            parameter can force AxographIO to put all signals into a single
            Segment. Default: False.

    Example:
        >>> import neo
        >>> r = neo.io.AxographIO(filename=filename)
        >>> blk = r.read_block(signal_group_mode='split-all')
        >>> display(blk)

        >>> # get signals
        >>> seg_index = 0  # episode number
        >>> sigs = [sig for sig in blk.segments[seg_index].analogsignals
        ...         if sig.name in channel_names]
        >>> display(sigs)

        >>> # get event markers (same for all segments/episodes)
        >>> ev = blk.segments[0].events[0]
        >>> print([ev for ev in zip(ev.times, ev.labels)])

        >>> # get interval bars (same for all segments/episodes)
        >>> ep = blk.segments[0].epochs[0]
        >>> print([ep for ep in zip(ep.times, ep.durations, ep.labels)])

        >>> # get notes
        >>> print(blk.annotations['notes'])
    """

    name = "AxographIO"
    description = "This IO reads .axgd/.axgx files created with AxoGraph"

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename="", force_single_segment=False):
        AxographRawIO.__init__(self, filename, force_single_segment)
        BaseFromRaw.__init__(self, filename)



import unittest
import os
from PIL import Image
import numpy as np
import shutil
from neo.io.tiffio import TiffIO
import quantities as pq


class TestTiffIO(unittest.TestCase):

    def test_read_group_of_tiff_grayscale(self):
        directory = "test_tiff"
        if not os.path.exists(directory):
            os.makedirs(directory)
        # directory is live
        img = []
        for picture in range(10):
            img.append([])
            for y in range(50):
                img[picture].append([])
                for x in range(50):
                    img[picture][y].append(x)
        img = np.array(img, dtype=float)
        for image in range(10):
            Image.fromarray(img[image]).save(directory + "/tiff_exemple" + str(image) + ".tif")

        ioclass = TiffIO(
            directory_path=directory, units="V", sampling_rate=1.0 * pq.Hz, spatial_scale=1.0 * pq.micrometer
        )
        blck = ioclass.read_block()
        self.assertEqual(len(blck.segments), 1)
        self.assertEqual(len(blck.segments[0].imagesequences), 1)
        self.assertEqual(blck.segments[0].imagesequences[0].any(), img.any())
        self.assertEqual(blck.segments[0].imagesequences[0].sampling_rate, 1.0 * pq.Hz)
        self.assertEqual(blck.segments[0].imagesequences[0].spatial_scale, 1.0 * pq.micrometer)

        # end of directory
        shutil.rmtree(directory)


if __name__ == "__main__":
    unittest.main()



"""
NeuroshareIO is a wrap with ctypes of neuroshare DLLs.
Neuroshare is a C API for reading neural data.
Neuroshare also provides a Matlab and a Python API on top of that.

Neuroshare is an open source API but each dll is provided directly by the vendor.
The neo user have to download separtatly the dll on neurosharewebsite:
http://neuroshare.sourceforge.net/

For some vendors (Spike2/CED , Clampfit/Abf, ...), neo.io also provides pure Python
Neo users you should prefer them of course :)

Supported : Read

Author: sgarcia

"""

import sys
import ctypes
import os
import warnings

import numpy as np
import quantities as pq

from neo.io.baseio import BaseIO
from neo.core import Segment, AnalogSignal, SpikeTrain, Event

ns_OK = 0  # Function successful
ns_LIBERROR = -1  # Generic linked library error
ns_TYPEERROR = -2  # Library unable to open file type
ns_FILEERROR = -3  # File access or read error
ns_BADFILE = -4  # Invalid file handle passed to function
ns_BADENTITY = -5  # Invalid or inappropriate entity identifier specified
ns_BADSOURCE = -6  # Invalid source identifier specified
ns_BADINDEX = -7  # Invalid entity index specified


class NeuroshareError(Exception):
    def __init__(self, lib, errno):
        self.lib = lib
        self.errno = errno
        pszMsgBuffer = ctypes.create_string_buffer(256)
        self.lib.ns_GetLastErrorMsg(pszMsgBuffer, ctypes.c_uint32(256))
        errstr = f"{errno}: {pszMsgBuffer.value}"
        Exception.__init__(self, errstr)


class DllWithError:
    def __init__(self, lib):
        self.lib = lib

    def __getattr__(self, attr):
        f = getattr(self.lib, attr)
        return self.decorate_with_error(f)

    def decorate_with_error(self, f):
        def func_with_error(*args):
            errno = f(*args)
            if errno != ns_OK:
                raise NeuroshareError(self.lib, errno)
            return errno

        return func_with_error


class NeurosharectypesIO(BaseIO):
    """
    Class for reading file trougth neuroshare API.
    The user need the DLLs in the path of the file format.

    Usage:
        >>> from neo import io
        >>> r = io.NeuroshareIO(filename='a_file', dllname=the_name_of_dll)
        >>> seg = r.read_segment(import_neuroshare_segment=True)
        >>> print seg.analogsignals        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        [<AnalogSignal(array([ -1.77246094e+02,  -2.24707031e+02,  -2.66015625e+02,
        ...
        >>> print seg.spiketrains
        []
        >>> print seg.events
        [<EventArray: 1@1.12890625 s, 1@2.02734375 s, 1@3.82421875 s>]

    Note:
        neuroshare.ns_ENTITY_EVENT: are converted to neo.EventArray
        neuroshare.ns_ENTITY_ANALOG: are converted to neo.AnalogSignal
        neuroshare.ns_ENTITY_NEURALEVENT: are converted to neo.SpikeTrain

        neuroshare.ns_ENTITY_SEGMENT: is something between serie of small AnalogSignal
                                        and Spiketrain with associated waveforms.
                                        It is arbitrarily converted as SpikeTrain.


    """

    is_readable = True
    is_writable = False

    supported_objects = [Segment, AnalogSignal, Event, SpikeTrain]
    readable_objects = [Segment]
    writeable_objects = []

    has_header = False
    is_streameable = False

    read_params = {Segment: []}
    write_params = None

    name = "neuroshare"
    extensions = ["mcd"]
    mode = "file"

    def __init__(self, filename="", dllname=""):
        """
        Arguments:
            filename: the file to read
            ddlname: the name of neuroshare dll to be used for this file
        """
        BaseIO.__init__(self)
        self.dllname = dllname
        self.filename = str(filename)

    def read_segment(self, import_neuroshare_segment=True, lazy=False):
        """
        Arguments:
            import_neuroshare_segment: import neuroshare segment as SpikeTrain
            with associated waveforms or not imported at all.
        """
        assert not lazy, "Do not support lazy"

        seg = Segment(
            file_origin=os.path.basename(self.filename),
        )

        if self.dllname == "":
            warnings.warn("No neuroshare dll provided. Can not load data.")
            return Segment()

        if sys.platform.startswith("win"):
            neuroshare = ctypes.windll.LoadLibrary(self.dllname)
        elif sys.platform.startswith("linux"):
            neuroshare = ctypes.cdll.LoadLibrary(self.dllname)
        neuroshare = DllWithError(neuroshare)

        # elif sys.platform.startswith('darwin'):

        # API version
        info = ns_LIBRARYINFO()
        neuroshare.ns_GetLibraryInfo(ctypes.byref(info), ctypes.sizeof(info))
        seg.annotate(neuroshare_version=str(info.dwAPIVersionMaj) + "." + str(info.dwAPIVersionMin))

        # open file
        hFile = ctypes.c_uint32(0)
        neuroshare.ns_OpenFile(ctypes.c_char_p(bytes(self.filename, "utf-8")), ctypes.byref(hFile))
        fileinfo = ns_FILEINFO()
        neuroshare.ns_GetFileInfo(hFile, ctypes.byref(fileinfo), ctypes.sizeof(fileinfo))

        # read all entities
        for dwEntityID in range(fileinfo.dwEntityCount):
            entityInfo = ns_ENTITYINFO()
            neuroshare.ns_GetEntityInfo(hFile, dwEntityID, ctypes.byref(entityInfo), ctypes.sizeof(entityInfo))

            # EVENT
            if entity_types[entityInfo.dwEntityType] == "ns_ENTITY_EVENT":
                pEventInfo = ns_EVENTINFO()
                neuroshare.ns_GetEventInfo(hFile, dwEntityID, ctypes.byref(pEventInfo), ctypes.sizeof(pEventInfo))

                if pEventInfo.dwEventType == 0:  # TEXT
                    pData = ctypes.create_string_buffer(pEventInfo.dwMaxDataLength)
                elif pEventInfo.dwEventType == 1:  # CVS
                    pData = ctypes.create_string_buffer(pEventInfo.dwMaxDataLength)
                elif pEventInfo.dwEventType == 2:  # 8bit
                    pData = ctypes.c_byte(0)
                elif pEventInfo.dwEventType == 3:  # 16bit
                    pData = ctypes.c_int16(0)
                elif pEventInfo.dwEventType == 4:  # 32bit
                    pData = ctypes.c_int32(0)
                pdTimeStamp = ctypes.c_double(0.0)
                pdwDataRetSize = ctypes.c_uint32(0)

                times = []
                labels = []
                for dwIndex in range(entityInfo.dwItemCount):
                    neuroshare.ns_GetEventData(
                        hFile,
                        dwEntityID,
                        dwIndex,
                        ctypes.byref(pdTimeStamp),
                        ctypes.byref(pData),
                        ctypes.sizeof(pData),
                        ctypes.byref(pdwDataRetSize),
                    )
                    times.append(pdTimeStamp.value)
                    labels.append(str(pData.value))
                times = times * pq.s
                labels = np.array(labels, dtype="U")

                ea = Event(name=str(entityInfo.szEntityLabel), times=times, labels=labels)
                seg.events.append(ea)

            # analog
            if entity_types[entityInfo.dwEntityType] == "ns_ENTITY_ANALOG":
                pAnalogInfo = ns_ANALOGINFO()

                neuroshare.ns_GetAnalogInfo(hFile, dwEntityID, ctypes.byref(pAnalogInfo), ctypes.sizeof(pAnalogInfo))
                dwIndexCount = entityInfo.dwItemCount

                pdwContCount = ctypes.c_uint32(0)
                pData = np.zeros((entityInfo.dwItemCount,), dtype="float64")
                total_read = 0
                while total_read < entityInfo.dwItemCount:
                    dwStartIndex = ctypes.c_uint32(total_read)
                    dwStopIndex = ctypes.c_uint32(entityInfo.dwItemCount - total_read)

                    neuroshare.ns_GetAnalogData(
                        hFile,
                        dwEntityID,
                        dwStartIndex,
                        dwStopIndex,
                        ctypes.byref(pdwContCount),
                        pData[total_read:].ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                    )
                    total_read += pdwContCount.value

                try:
                    signal = pq.Quantity(pData, units=pAnalogInfo.szUnits.decode(), copy=False)
                    unit_annotation = None
                except LookupError:
                    signal = pq.Quantity(pData, units="dimensionless", copy=False)
                    unit_annotation = pAnalogInfo.szUnits.decode()

                # t_start
                dwIndex = 0
                pdTime = ctypes.c_double(0)
                neuroshare.ns_GetTimeByIndex(hFile, dwEntityID, dwIndex, ctypes.byref(pdTime))

                anaSig = AnalogSignal(
                    signal,
                    sampling_rate=pAnalogInfo.dSampleRate * pq.Hz,
                    t_start=pdTime.value * pq.s,
                    name=str(entityInfo.szEntityLabel),
                )
                anaSig.annotate(probe_info=str(pAnalogInfo.szProbeInfo))
                if unit_annotation is not None:
                    anaSig.annotate(units=unit_annotation)
                seg.analogsignals.append(anaSig)

            # segment
            if entity_types[entityInfo.dwEntityType] == "ns_ENTITY_SEGMENT" and import_neuroshare_segment:

                pdwSegmentInfo = ns_SEGMENTINFO()
                if not str(entityInfo.szEntityLabel).startswith("spks"):
                    continue

                neuroshare.ns_GetSegmentInfo(
                    hFile, dwEntityID, ctypes.byref(pdwSegmentInfo), ctypes.sizeof(pdwSegmentInfo)
                )
                nsource = pdwSegmentInfo.dwSourceCount

                pszMsgBuffer = ctypes.create_string_buffer(" " * 256)
                neuroshare.ns_GetLastErrorMsg(ctypes.byref(pszMsgBuffer), 256)

                for dwSourceID in range(pdwSegmentInfo.dwSourceCount):
                    pSourceInfo = ns_SEGSOURCEINFO()
                    neuroshare.ns_GetSegmentSourceInfo(
                        hFile, dwEntityID, dwSourceID, ctypes.byref(pSourceInfo), ctypes.sizeof(pSourceInfo)
                    )

                pdTimeStamp = ctypes.c_double(0.0)
                dwDataBufferSize = pdwSegmentInfo.dwMaxSampleCount * pdwSegmentInfo.dwSourceCount
                pData = np.zeros((dwDataBufferSize), dtype="float64")
                pdwSampleCount = ctypes.c_uint32(0)
                pdwUnitID = ctypes.c_uint32(0)

                nsample = int(dwDataBufferSize)
                times = np.empty((entityInfo.dwItemCount), dtype="f")
                waveforms = np.empty((entityInfo.dwItemCount, nsource, nsample), dtype="f")
                for dwIndex in range(entityInfo.dwItemCount):
                    neuroshare.ns_GetSegmentData(
                        hFile,
                        dwEntityID,
                        dwIndex,
                        ctypes.byref(pdTimeStamp),
                        pData.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                        dwDataBufferSize * 8,
                        ctypes.byref(pdwSampleCount),
                        ctypes.byref(pdwUnitID),
                    )

                    times[dwIndex] = pdTimeStamp.value
                    waveforms[dwIndex, :, :] = pData[: nsample * nsource].reshape(nsample, nsource).transpose()

                sptr = SpikeTrain(
                    times=pq.Quantity(times, units="s", copy=False),
                    t_stop=times.max(),
                    waveforms=pq.Quantity(waveforms, units=str(pdwSegmentInfo.szUnits), copy=False),
                    left_sweep=nsample / 2.0 / float(pdwSegmentInfo.dSampleRate) * pq.s,
                    sampling_rate=float(pdwSegmentInfo.dSampleRate) * pq.Hz,
                    name=str(entityInfo.szEntityLabel),
                )
                seg.spiketrains.append(sptr)

            # neuralevent
            if entity_types[entityInfo.dwEntityType] == "ns_ENTITY_NEURALEVENT":
                pNeuralInfo = ns_NEURALINFO()
                neuroshare.ns_GetNeuralInfo(hFile, dwEntityID, ctypes.byref(pNeuralInfo), ctypes.sizeof(pNeuralInfo))

                pData = np.zeros((entityInfo.dwItemCount,), dtype="float64")
                dwStartIndex = 0
                dwIndexCount = entityInfo.dwItemCount
                neuroshare.ns_GetNeuralData(
                    hFile, dwEntityID, dwStartIndex, dwIndexCount, pData.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
                )
                times = pData * pq.s
                t_stop = times.max()

                sptr = SpikeTrain(
                    times,
                    t_stop=t_stop,
                    name=str(entityInfo.szEntityLabel),
                )
                seg.spiketrains.append(sptr)

        # close
        neuroshare.ns_CloseFile(hFile)

        seg.check_relationships()
        return seg


# neuroshare structures
class ns_FILEDESC(ctypes.Structure):
    _fields_ = [
        ("szDescription", ctypes.c_char * 32),
        ("szExtension", ctypes.c_char * 8),
        ("szMacCodes", ctypes.c_char * 8),
        ("szMagicCode", ctypes.c_char * 16),
    ]


class ns_LIBRARYINFO(ctypes.Structure):
    _fields_ = [
        ("dwLibVersionMaj", ctypes.c_uint32),
        ("dwLibVersionMin", ctypes.c_uint32),
        ("dwAPIVersionMaj", ctypes.c_uint32),
        ("dwAPIVersionMin", ctypes.c_uint32),
        ("szDescription", ctypes.c_char * 64),
        ("szCreator", ctypes.c_char * 64),
        ("dwTime_Year", ctypes.c_uint32),
        ("dwTime_Month", ctypes.c_uint32),
        ("dwTime_Day", ctypes.c_uint32),
        ("dwFlags", ctypes.c_uint32),
        ("dwMaxFiles", ctypes.c_uint32),
        ("dwFileDescCount", ctypes.c_uint32),
        ("FileDesc", ns_FILEDESC * 16),
    ]


class ns_FILEINFO(ctypes.Structure):
    _fields_ = [
        ("szFileType", ctypes.c_char * 32),
        ("dwEntityCount", ctypes.c_uint32),
        ("dTimeStampResolution", ctypes.c_double),
        ("dTimeSpan", ctypes.c_double),
        ("szAppName", ctypes.c_char * 64),
        ("dwTime_Year", ctypes.c_uint32),
        ("dwTime_Month", ctypes.c_uint32),
        ("dwReserved", ctypes.c_uint32),
        ("dwTime_Day", ctypes.c_uint32),
        ("dwTime_Hour", ctypes.c_uint32),
        ("dwTime_Min", ctypes.c_uint32),
        ("dwTime_Sec", ctypes.c_uint32),
        ("dwTime_MilliSec", ctypes.c_uint32),
        ("szFileComment", ctypes.c_char * 256),
    ]


class ns_ENTITYINFO(ctypes.Structure):
    _fields_ = [
        ("szEntityLabel", ctypes.c_char * 32),
        ("dwEntityType", ctypes.c_uint32),
        ("dwItemCount", ctypes.c_uint32),
    ]


entity_types = {
    0: "ns_ENTITY_UNKNOWN",
    1: "ns_ENTITY_EVENT",
    2: "ns_ENTITY_ANALOG",
    3: "ns_ENTITY_SEGMENT",
    4: "ns_ENTITY_NEURALEVENT",
}


class ns_EVENTINFO(ctypes.Structure):
    _fields_ = [
        ("dwEventType", ctypes.c_uint32),
        ("dwMinDataLength", ctypes.c_uint32),
        ("dwMaxDataLength", ctypes.c_uint32),
        ("szCSVDesc", ctypes.c_char * 128),
    ]


class ns_ANALOGINFO(ctypes.Structure):
    _fields_ = [
        ("dSampleRate", ctypes.c_double),
        ("dMinVal", ctypes.c_double),
        ("dMaxVal", ctypes.c_double),
        ("szUnits", ctypes.c_char * 16),
        ("dResolution", ctypes.c_double),
        ("dLocationX", ctypes.c_double),
        ("dLocationY", ctypes.c_double),
        ("dLocationZ", ctypes.c_double),
        ("dLocationUser", ctypes.c_double),
        ("dHighFreqCorner", ctypes.c_double),
        ("dwHighFreqOrder", ctypes.c_uint32),
        ("szHighFilterType", ctypes.c_char * 16),
        ("dLowFreqCorner", ctypes.c_double),
        ("dwLowFreqOrder", ctypes.c_uint32),
        ("szLowFilterType", ctypes.c_char * 16),
        ("szProbeInfo", ctypes.c_char * 128),
    ]


class ns_SEGMENTINFO(ctypes.Structure):
    _fields_ = [
        ("dwSourceCount", ctypes.c_uint32),
        ("dwMinSampleCount", ctypes.c_uint32),
        ("dwMaxSampleCount", ctypes.c_uint32),
        ("dSampleRate", ctypes.c_double),
        ("szUnits", ctypes.c_char * 32),
    ]


class ns_SEGSOURCEINFO(ctypes.Structure):
    _fields_ = [
        ("dMinVal", ctypes.c_double),
        ("dMaxVal", ctypes.c_double),
        ("dResolution", ctypes.c_double),
        ("dSubSampleShift", ctypes.c_double),
        ("dLocationX", ctypes.c_double),
        ("dLocationY", ctypes.c_double),
        ("dLocationZ", ctypes.c_double),
        ("dLocationUser", ctypes.c_double),
        ("dHighFreqCorner", ctypes.c_double),
        ("dwHighFreqOrder", ctypes.c_uint32),
        ("szHighFilterType", ctypes.c_char * 16),
        ("dLowFreqCorner", ctypes.c_double),
        ("dwLowFreqOrder", ctypes.c_uint32),
        ("szLowFilterType", ctypes.c_char * 16),
        ("szProbeInfo", ctypes.c_char * 128),
    ]


class ns_NEURALINFO(ctypes.Structure):
    _fields_ = [
        ("dwSourceEntityID", ctypes.c_uint32),
        ("dwSourceUnitID", ctypes.c_uint32),
        ("szProbeInfo", ctypes.c_char * 128),
    ]



"""
:mod:`neo.io` provides classes for reading and/or writing
electrophysiological data files.

Note that if the package dependency is not satisfied for one io, it does not
raise an error but a warning.

:attr:`neo.io.iolist` provides a list of successfully imported io classes.

Functions:

.. autofunction:: neo.io.get_io
.. autofunction:: neo.io.list_candidate_ios


Classes:

* :attr:`AlphaOmegaIO`
* :attr:`AsciiImageIO`
* :attr:`AsciiSignalIO`
* :attr:`AsciiSpikeTrainIO`
* :attr:`AxographIO`
* :attr:`AxonaIO`
* :attr:`AxonIO`
* :attr:`BCI2000IO`
* :attr:`BiocamIO`
* :attr:`BlackrockIO`
* :attr:`BlkIO`
* :attr:`BrainVisionIO`
* :attr:`BrainwareDamIO`
* :attr:`BrainwareF32IO`
* :attr:`BrainwareSrcIO`
* :attr:`CedIO`
* :attr:`EDFIO`
* :attr:`ElanIO`
* :attr:`IgorIO`
* :attr:`IntanIO`
* :attr:`MEArecIO`
* :attr:`KlustaKwikIO`
* :attr:`KwikIO`
* :attr:`MaxwellIO`
* :attr:`MedIO`
* :attr:`MicromedIO`
* :attr:`NeoMatlabIO`
* :attr:`NestIO`
* :attr:`NeuralynxIO`
* :attr:`NeuroExplorerIO`
* :attr:`NeuroScopeIO`
* :attr:`NeuroshareIO`
* :attr:`NixIO`
* :attr:`NWBIO`
* :attr:`OpenEphysIO`
* :attr:`OpenEphysBinaryIO`
* :attr:`PhyIO`
* :attr:`PickleIO`
* :attr:`PlexonIO`
* :attr:`Plexon2IO`
* :attr:`RawBinarySignalIO`
* :attr:`RawMCSIO`
* :attr:`Spike2IO`
* :attr:`SpikeGadgetsIO`
* :attr:`SpikeGLXIO`
* :attr:`StimfitIO`
* :attr:`TdtIO`
* :attr:`TiffIO`
* :attr:`WinEdrIO`
* :attr:`WinWcpIO`


.. autoclass:: neo.io.AlphaOmegaIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.AsciiImageIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.AsciiSignalIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.AsciiSpikeTrainIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.AxographIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.AxonaIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.AxonIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BCI2000IO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BiocamIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BlackrockIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BlkIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BrainVisionIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BrainwareDamIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BrainwareF32IO

    .. autoattribute:: extensions

.. autoclass:: neo.io.BrainwareSrcIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.CedIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.EDFIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.ElanIO

    .. autoattribute:: extensions

.. .. autoclass:: neo.io.ElphyIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.IgorIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.IntanIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.KlustaKwikIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.KwikIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.MEArecIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.MaxwellIO

    .. autoattribute:: extensions
    
.. autoclass:: neo.io.MedIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.MicromedIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NeoMatlabIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NestIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NeuralynxIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NeuroExplorerIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NeuroScopeIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NeuroshareIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NixIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.NWBIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.OpenEphysIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.OpenEphysBinaryIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.PhyIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.PickleIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.PlexonIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.Plexon2IO

    .. autoattribute:: extensions

.. autoclass:: neo.io.RawBinarySignalIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.RawMCSIO

    .. autoattribute:: extensions

.. autoclass:: Spike2IO

    .. autoattribute:: extensions

.. autoclass:: SpikeGadgetsIO

    .. autoattribute:: extensions

.. autoclass:: SpikeGLXIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.StimfitIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.TdtIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.TiffIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.WinEdrIO

    .. autoattribute:: extensions

.. autoclass:: neo.io.WinWcpIO

    .. autoattribute:: extensions

"""

import pathlib
from collections import Counter

# try to import the neuroshare library.
# if it is present, use the neuroshareapiio to load neuroshare files
# if it is not present, use the neurosharectypesio to load files
try:
    import neuroshare as ns
except ModuleNotFoundError as err:
    from neo.io.neurosharectypesio import NeurosharectypesIO as NeuroshareIO

    # print("\n neuroshare library not found, loading data with ctypes" )
    # print("\n to use the API be sure to install the library found at:")
    # print("\n www.http://pythonhosted.org/neuroshare/")

else:
    from neo.io.neuroshareapiio import NeuroshareapiIO as NeuroshareIO

    # print("neuroshare library successfully imported")
    # print("\n loading with API...")

from neo.io.alphaomegaio import AlphaOmegaIO
from neo.io.asciiimageio import AsciiImageIO
from neo.io.asciisignalio import AsciiSignalIO
from neo.io.asciispiketrainio import AsciiSpikeTrainIO
from neo.io.axographio import AxographIO
from neo.io.axonaio import AxonaIO
from neo.io.axonio import AxonIO
from neo.io.biocamio import BiocamIO
from neo.io.blackrockio import BlackrockIO
from neo.io.blkio import BlkIO
from neo.io.bci2000io import BCI2000IO
from neo.io.brainvisionio import BrainVisionIO
from neo.io.brainwaredamio import BrainwareDamIO
from neo.io.brainwaref32io import BrainwareF32IO
from neo.io.brainwaresrcio import BrainwareSrcIO
from neo.io.cedio import CedIO
from neo.io.edfio import EDFIO
from neo.io.elanio import ElanIO
from neo.io.elphyio import ElphyIO
from neo.io.exampleio import ExampleIO
from neo.io.igorproio import IgorIO
from neo.io.intanio import IntanIO
from neo.io.klustakwikio import KlustaKwikIO
from neo.io.kwikio import KwikIO
from neo.io.mearecio import MEArecIO
from neo.io.maxwellio import MaxwellIO
from neo.io.medio import MedIO
from neo.io.micromedio import MicromedIO
from neo.io.neomatlabio import NeoMatlabIO
from neo.io.nestio import NestIO
from neo.io.neuralynxio import NeuralynxIO
from neo.io.neuroexplorerio import NeuroExplorerIO
from neo.io.neuroscopeio import NeuroScopeIO
from neo.io.nixio import NixIO
from neo.io.nixio_fr import NixIO as NixIOFr
from neo.io.nwbio import NWBIO
from neo.io.openephysio import OpenEphysIO
from neo.io.openephysbinaryio import OpenEphysBinaryIO
from neo.io.phyio import PhyIO
from neo.io.pickleio import PickleIO
from neo.io.plexonio import PlexonIO
from neo.io.plexon2io import Plexon2IO
from neo.io.rawbinarysignalio import RawBinarySignalIO
from neo.io.rawmcsio import RawMCSIO
from neo.io.spike2io import Spike2IO
from neo.io.spikegadgetsio import SpikeGadgetsIO
from neo.io.spikeglxio import SpikeGLXIO
from neo.io.stimfitio import StimfitIO
from neo.io.tdtio import TdtIO
from neo.io.tiffio import TiffIO
from neo.io.winedrio import WinEdrIO
from neo.io.winwcpio import WinWcpIO

iolist = [
    AlphaOmegaIO,
    AsciiImageIO,
    AsciiSignalIO,
    AsciiSpikeTrainIO,
    AxographIO,
    AxonaIO,
    AxonIO,
    BCI2000IO,
    BiocamIO,
    BlackrockIO,
    BlkIO,
    BrainVisionIO,
    BrainwareDamIO,
    BrainwareF32IO,
    BrainwareSrcIO,
    CedIO,
    EDFIO,
    ElanIO,
    ElphyIO,
    ExampleIO,
    IgorIO,
    IntanIO,
    KlustaKwikIO,
    KwikIO,
    MEArecIO,
    MaxwellIO,
    MedIO,
    MicromedIO,
    NixIO,
    NixIOFr,
    NeoMatlabIO,
    NestIO,
    NeuralynxIO,
    NeuroExplorerIO,
    NeuroScopeIO,
    NeuroshareIO,
    NWBIO,
    OpenEphysIO,
    OpenEphysBinaryIO,
    PhyIO,
    PickleIO,
    PlexonIO,
    Plexon2IO,
    RawBinarySignalIO,
    RawMCSIO,
    Spike2IO,
    SpikeGadgetsIO,
    SpikeGLXIO,
    StimfitIO,
    TdtIO,
    TiffIO,
    WinEdrIO,
    WinWcpIO,
]

# for each supported extension list the ios supporting it
io_by_extension = {}
for current_io in iolist:  # do not use `io` as variable name here as this overwrites the module io
    for extension in current_io.extensions:
        extension = extension.lower()
        # extension handling should not be case sensitive
        io_by_extension.setdefault(extension, []).append(current_io)


def get_io(file_or_folder, *args, **kwargs):
    """
    Return a Neo IO instance, guessing the type based on the filename suffix.
    """
    ios = list_candidate_ios(file_or_folder)
    for io in ios:
        try:
            return io(file_or_folder, *args, **kwargs)
        except:
            continue

    raise IOError(f"Could not identify IO for {file_or_folder}")


def list_candidate_ios(file_or_folder, ignore_patterns=["*.ini", "README.txt", "README.md"]):
    """
    Identify neo IO that can potentially load data in the file or folder

    Parameters
    ----------
    file_or_folder (str, pathlib.Path)
        Path to the file or folder to load
    ignore_patterns (list)
        List of patterns to ignore when scanning for known formats. See pathlib.PurePath.match().
        Default: ['ini']

    Returns
    -------
    list
        List of neo io classes that are associated with the file extensions detected
    """
    file_or_folder = pathlib.Path(file_or_folder)

    if file_or_folder.is_file():
        suffix = file_or_folder.suffix[1:].lower()
        if suffix not in io_by_extension:
            raise ValueError(f"{suffix} is not a supported format of any IO.")
        return io_by_extension[suffix]

    elif file_or_folder.is_dir():
        # scan files in folder to determine io type
        filenames = [f for f in file_or_folder.glob("**/*") if f.is_file()]
        # keep only relevant filenames
        filenames = [f for f in filenames if f.suffix and not any([f.match(p) for p in ignore_patterns])]

        # if no files are found in the folder, check subfolders
        # this is necessary for nested-folder based formats like spikeglx
        if not filenames:
            filenames = [f for f in file_or_folder.glob("**/*") if f.is_file()]
            # keep only relevant filenames
            filenames = [f for f in filenames if f.suffix and not any([f.match(p) for p in ignore_patterns])]

    # if only file prefix was provided, e.g /mydatafolder/session1-
    # to select all files sharing the `session1-` prefix
    elif file_or_folder.parent.exists():
        filenames = list(file_or_folder.parent.glob(file_or_folder.name + "*"))
        # if filenames empty and suffix is provided then non-existent file
        # may be written in current dir. So run check for io
        if len(filenames) == 0 and file_or_folder.suffix:
            suffix = file_or_folder.suffix[1:].lower()
            if suffix not in io_by_extension:
                raise ValueError(f"{suffix} is not a supported format of any IO.")
            return io_by_extension[suffix]

    # If non-existent file in non-existent dir is given check if this
    # structure could be created with an io writing the file
    elif file_or_folder.suffix:
        suffix = file_or_folder.suffix[1:].lower()
        if suffix not in io_by_extension:
            raise ValueError(f"{suffix} is not a supported format of any IO.")
        return io_by_extension[suffix]

    else:
        raise ValueError(f"{file_or_folder} does not contain data files of a supported format")

    # find the io that fits the best with the files contained in the folder
    potential_ios = []
    for filename in filenames:
        for suffix in filename.suffixes:
            suffix = suffix[1:].lower()
            if suffix in io_by_extension:
                potential_ios.extend(io_by_extension[suffix])

    if not potential_ios:
        raise ValueError(f"Could not determine IO to load {file_or_folder}")

    # return ios ordered by number of files supported
    counter = Counter(potential_ios).most_common()
    return [io for io, count in counter]



import unittest

import os
import numpy as np

from neo.rawio.neuralynxrawio.neuralynxrawio import NeuralynxRawIO
from neo.rawio.neuralynxrawio.nlxheader import NlxHeader
from neo.rawio.neuralynxrawio.ncssections import NcsSection, NcsSections, NcsSectionsFactory
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

import logging

logging.getLogger().setLevel(logging.INFO)


class TestNeuralynxRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = NeuralynxRawIO
    entities_to_download = ["neuralynx"]
    entities_to_test = [
        "neuralynx/BML/original_data",
        "neuralynx/BML_unfilledsplit/original_data",
        "neuralynx/Cheetah_v1.1.0/original_data",
        "neuralynx/Cheetah_v4.0.2/original_data",
        "neuralynx/Cheetah_v5.4.0/original_data",
        "neuralynx/Cheetah_v5.5.1/original_data",
        "neuralynx/Cheetah_v5.6.3/original_data",
        "neuralynx/Cheetah_v5.7.4/original_data",
        "neuralynx/Cheetah_v6.3.2/incomplete_blocks",
    ]

    def test_scan_ncs_files(self):

        # Test BML style of Ncs files, similar to PRE4 but with fractional frequency
        # in the header and fractional microsPerSamp, which is then rounded as appropriate
        # in each record.
        rawio = NeuralynxRawIO(self.get_local_path("neuralynx/BML/original_data"))
        rawio.parse_header()
        # test values here from direct inspection of .ncs files
        self.assertEqual(rawio._nb_segment, 1)
        self.assertListEqual(rawio._timestamp_limits, [(0, 192000)])
        self.assertEqual(rawio._sigs_length[0][("unknown", "1")], 4608)
        self.assertListEqual(rawio._signal_limits, [(0, 0.192)])
        self.assertEqual(len(rawio._sigs_memmaps), 1)

        # Test Cheetah 4.0.2, which is PRE4 type with frequency in header and
        # no microsPerSamp. Number of microseconds per sample in file is inverse of
        # sampling frequency in header trucated to microseconds.
        rawio = NeuralynxRawIO(self.get_local_path("neuralynx/Cheetah_v4.0.2/original_data"))
        rawio.parse_header()
        # test values here from direct inspection of .ncs files
        self.assertEqual(rawio._nb_segment, 1)
        self.assertEqual(rawio.signal_streams_count(), 1)
        self.assertListEqual(rawio._timestamp_limits, [(266982936, 267162136)])
        self.assertEqual(rawio._sigs_length[0][("unknown", "13")], 5120)
        self.assertListEqual(rawio._signal_limits, [(266.982936, 267.162136)])
        self.assertEqual(len(rawio._sigs_memmaps), 1)

        # Test Cheetah 5.5.1, which is DigitalLynxSX and has two blocks of records
        # with a fairly large gap.
        rawio = NeuralynxRawIO(self.get_local_path("neuralynx/Cheetah_v5.5.1/original_data"))
        rawio.parse_header()
        # test values here from direct inspection of .ncs files
        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(rawio.signal_streams_count(), 1)
        self.assertListEqual(rawio._timestamp_limits, [(26122557633, 26162525633), (26366360633, 26379704633)])
        self.assertDictEqual(rawio._sigs_length[0], {("Tet3a", "8"): 1278976, ("Tet3b", "9"): 1278976})
        self.assertDictEqual(rawio._sigs_length[1], {("Tet3a", "8"): 427008, ("Tet3b", "9"): 427008})
        self.assertListEqual(rawio._signal_limits, [(26122.557633, 26162.525633), (26366.360633, 26379.704633)])
        self.assertEqual(len(rawio._sigs_memmaps), 2)  # check only that there are 2 memmaps

        # Test Cheetah 6.3.2, the incomplete_blocks test. This is a DigitalLynxSX with
        # three blocks of records. Gaps are on the order of 60 microseconds or so.
        rawio = NeuralynxRawIO(self.get_local_path("neuralynx/Cheetah_v6.3.2/incomplete_blocks"))
        rawio.parse_header()
        # test values here from direct inspection of .ncs file, except for 3rd block
        # t_stop, which is extended due to events past the last block of ncs records.
        self.assertEqual(rawio._nb_segment, 3)
        self.assertEqual(rawio.signal_streams_count(), 1)
        self.assertListEqual(
            rawio._timestamp_limits, [(8408806811, 8427831990), (8427832053, 8487768498), (8487768561, 8515816549)]
        )
        self.assertDictEqual(rawio._sigs_length[0], {("CSC1", "48"): 608806})
        self.assertDictEqual(rawio._sigs_length[1], {("CSC1", "48"): 1917967})
        self.assertDictEqual(rawio._sigs_length[2], {("CSC1", "48"): 897536})
        self.assertListEqual(
            rawio._signal_limits, [(8408.806811, 8427.831990), (8427.832053, 8487.768498), (8487.768561, 8515.816549)]
        )
        self.assertEqual(len(rawio._sigs_memmaps), 3)  # check that there are only 3 memmaps

        # Test Cheetah 6.4.1, with different sampling rates across ncs files.
        rawio = NeuralynxRawIO(self.get_local_path("neuralynx/Cheetah_v6.4.1dev/original_data"))
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 1)
        seg_idx = 0

        self.assertEqual(rawio.signal_streams_count(), 3)
        self.assertListEqual(rawio._timestamp_limits, [(1614363777825263, 1614363778481169)])
        self.assertDictEqual(
            rawio._sigs_length[seg_idx], {("CSC1", "26"): 15872, ("LFP4", "41"): 1024, ("WE1", "33"): 512}
        )
        self.assertListEqual(rawio._signal_limits, [(1614363777.825263, 1614363778.481169)])
        # check that there are only 3 memmaps
        self.assertEqual(len(rawio._sigs_memmaps[seg_idx]), 3)

    def test_include_filenames(self):
        """
        Tests include_filenames with only one file
        """

        # test single analog signal channel
        fname = self.get_local_path("neuralynx/Cheetah_v5.6.3/original_data/CSC1.ncs")
        dirname, filename = os.path.split(fname)
        rawio = NeuralynxRawIO(dirname=dirname, include_filenames=filename)
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(len(rawio.ncs_filenames), 1)
        self.assertEqual(len(rawio.nev_filenames), 0)
        sigHdrs = rawio.header["signal_channels"]
        self.assertEqual(sigHdrs.size, 1)
        self.assertEqual(sigHdrs[0][0], "CSC1")
        self.assertEqual(sigHdrs[0][1], "58")
        self.assertEqual(len(rawio.header["spike_channels"]), 0)
        self.assertEqual(len(rawio.header["event_channels"]), 0)

        # test one single electrode channel
        fname = self.get_local_path("neuralynx/Cheetah_v5.5.1/original_data/STet3a.nse")
        dirname, filename = os.path.split(fname)
        rawio = NeuralynxRawIO(dirname=dirname, include_filenames=filename)
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 1)
        self.assertEqual(len(rawio.ncs_filenames), 0)
        self.assertEqual(len(rawio.nev_filenames), 0)
        seHdrs = rawio.header["spike_channels"]
        self.assertEqual(len(seHdrs), 1)
        self.assertEqual(seHdrs[0][0], "chSTet3a#8#0")
        self.assertEqual(seHdrs[0][1], "0")
        self.assertEqual(len(rawio.header["signal_channels"]), 0)
        self.assertEqual(len(rawio.header["event_channels"]), 0)

    def test_exclude_filenames(self):
        # exclude single ncs file from session
        dname = self.get_local_path("neuralynx/Cheetah_v5.6.3/original_data/")
        rawio = NeuralynxRawIO(dirname=dname, exclude_filenames="CSC2.ncs")
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(len(rawio.ncs_filenames), 1)
        self.assertEqual(len(rawio.nev_filenames), 1)
        sigHdrs = rawio.header["signal_channels"]
        self.assertEqual(sigHdrs.size, 1)
        self.assertEqual(sigHdrs[0][0], "CSC1")
        self.assertEqual(sigHdrs[0][1], "58")
        self.assertEqual(len(rawio.header["spike_channels"]), 8)
        self.assertEqual(len(rawio.header["event_channels"]), 2)

        # exclude multiple files from session
        rawio = NeuralynxRawIO(dirname=dname, exclude_filenames=["Events.nev", "CSC2.ncs"])
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(len(rawio.ncs_filenames), 1)
        self.assertEqual(len(rawio.nev_filenames), 0)
        sigHdrs = rawio.header["signal_channels"]
        self.assertEqual(sigHdrs.size, 1)
        self.assertEqual(sigHdrs[0][0], "CSC1")
        self.assertEqual(sigHdrs[0][1], "58")
        self.assertEqual(len(rawio.header["spike_channels"]), 8)
        self.assertEqual(len(rawio.header["event_channels"]), 0)


class TestNcsRecordingType(TestNeuralynxRawIO, unittest.TestCase):
    """
    Test of decoding of NlxHeader for type of recording.
    """

    entities_to_test = []

    ncsTypeTestFiles = [
        ("neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs", "PRE4"),
        ("neuralynx/Cheetah_v5.4.0/original_data/CSC5_trunc.Ncs", "DIGITALLYNX"),
        ("neuralynx/Cheetah_v5.5.1/original_data/STet3a.nse", "DIGITALLYNXSX"),
        ("neuralynx/Cheetah_v5.5.1/original_data/Tet3a.ncs", "DIGITALLYNXSX"),
        ("neuralynx/Cheetah_v5.6.3/original_data/CSC1.ncs", "DIGITALLYNXSX"),
        ("neuralynx/Cheetah_v5.6.3/original_data/TT1.ntt", "DIGITALLYNXSX"),
        ("neuralynx/Cheetah_v5.7.4/original_data/CSC1.ncs", "DIGITALLYNXSX"),
        ("neuralynx/Cheetah_v6.3.2/incomplete_blocks/CSC1_reduced.ncs", "DIGITALLYNXSX"),
        ("neuralynx/Pegasus_v2.1.1/Events_0008.nev", "ATLAS"),
    ]

    def test_recording_types(self):

        for typeTest in self.ncsTypeTestFiles:

            filename = self.get_local_path(typeTest[0])
            hdr = NlxHeader(filename)
            self.assertEqual(hdr.type_of_recording(), typeTest[1])


class TestNcsSectionsFactory(TestNeuralynxRawIO, unittest.TestCase):
    """
    Test building NcsBlocks for files of different revisions.
    """

    entities_to_test = []

    def test_ncsblocks_partial(self):
        filename = self.get_local_path("neuralynx/Cheetah_v6.3.2/incomplete_blocks/CSC1_reduced.ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        self.assertEqual(data0.shape[0], 6690)
        self.assertEqual(data0["timestamp"][6689], 8515800549)  # timestamp of last record

        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(nb.sampFreqUsed, 32000.012813673042)
        self.assertEqual(nb.microsPerSampUsed, 31.249987486652431)
        self.assertListEqual([blk.startRec for blk in nb.sects], [0, 1190, 4937])
        self.assertListEqual([blk.endRec for blk in nb.sects], [1189, 4936, 6689])

    def test_build_given_actual_frequency(self):

        # Test early files where the frequency listed in the header is
        # floor(1e6/(actual number of microseconds between samples)
        filename = self.get_local_path("neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        ncsBlocks = NcsSections()
        ncsBlocks.sampFreqUsed = 1 / (35e-6)
        ncsBlocks.microsPerSampUsed = 35

        ncsBlocks = NcsSectionsFactory._buildNcsSections(data0, ncsBlocks.sampFreqUsed)

        self.assertEqual(len(ncsBlocks.sects), 1)
        self.assertEqual(ncsBlocks.sects[0].startRec, 0)
        self.assertEqual(ncsBlocks.sects[0].endRec, 9)

    def test_build_using_header_and_scanning(self):

        # Test early files where the frequency listed in the header is
        # floor(1e6/(actual number of microseconds between samples)
        filename = self.get_local_path("neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs")
        hdr = NlxHeader(filename)
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)

        self.assertEqual(nb.sampFreqUsed, 1 / 35e-6)
        self.assertEqual(nb.microsPerSampUsed, 35)
        self.assertEqual(len(nb.sects), 1)
        self.assertEqual(nb.sects[0].startRec, 0)
        self.assertEqual(nb.sects[0].endRec, 9)

        # test Cheetah 5.5.1, which is DigitalLynxSX and has two blocks of records
        # with a fairly large gap
        filename = self.get_local_path("neuralynx/Cheetah_v5.5.1/original_data/Tet3a.ncs")
        hdr = NlxHeader(filename)
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(nb.sampFreqUsed, 32000)
        self.assertEqual(nb.microsPerSampUsed, 31.25)
        self.assertEqual(len(nb.sects), 2)
        self.assertListEqual([blk.startRec for blk in nb.sects], [0, 2498])
        self.assertListEqual([blk.endRec for blk in nb.sects], [2497, 3331])

    def test_block_start_and_end_times(self):
        # digitallynxsx version to exercise the _parseForMaxGap function with multiple blocks
        filename = self.get_local_path("neuralynx/Cheetah_v6.3.2/incomplete_blocks/CSC1_reduced.ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertListEqual([blk.startTime for blk in nb.sects], [8408806811, 8427832053, 8487768561])
        self.assertListEqual([blk.endTime for blk in nb.sects], [8427831990, 8487768497, 8515816549])

        # digitallynxsx with single block of records to exercise path in _buildForMaxGap
        filename = self.get_local_path("neuralynx/Cheetah_v1.1.0/original_data/CSC67_trunc.Ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(len(nb.sects), 1)
        self.assertEqual(nb.sects[0].startTime, 253293161778)
        self.assertEqual(nb.sects[0].endTime, 253293349278)

        # PRE4 version with single block of records to exercise path in _buildGivenActualFrequency
        filename = self.get_local_path("neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(len(nb.sects), 1)
        self.assertEqual(nb.sects[0].startTime, 266982936)
        self.assertEqual(nb.sects[0].endTime, 267162136)

        # BML style with two blocks of records and one partially filled record to exercise
        # _parseGivenActualFrequency
        filename = self.get_local_path("neuralynx/BML_unfilledsplit/original_data/unfilledSplitRecords.Ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(len(nb.sects), 2)
        self.assertListEqual([blk.startTime for blk in nb.sects], [1837623129, 6132625241])
        self.assertListEqual([blk.endTime for blk in nb.sects], [1837651009, 6132642649])

    def test_block_verify(self):
        # check that file verifies against itself for single block
        filename = self.get_local_path("neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs")
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        hdr0 = NlxHeader(filename)
        nb0 = NcsSectionsFactory.build_for_ncs_file(data0, hdr0)

        self.assertTrue(NcsSectionsFactory._verifySectionsStructure(data0, nb0))

        # check that fails against file with two blocks
        filename = self.get_local_path("neuralynx/BML_unfilledsplit/original_data/unfilledSplitRecords.Ncs")
        data1 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)
        hdr1 = NlxHeader(filename)
        nb1 = NcsSectionsFactory.build_for_ncs_file(data1, hdr1)

        self.assertFalse(NcsSectionsFactory._verifySectionsStructure(data1, nb0))

        # check that two blocks verify against self
        self.assertTrue(NcsSectionsFactory._verifySectionsStructure(data1, nb1))


class TestNcsSections(TestNeuralynxRawIO, unittest.TestCase):
    """
    Test building NcsBlocks for files of different revisions.
    """

    entities_to_test = []

    def test_equality(self):
        ns0 = NcsSections()
        ns1 = NcsSections()

        ns0.microsPerSampUsed = 1
        ns1.microsPerSampUsed = 1
        ns0.sampFreqUsed = 300
        ns1.sampFreqUsed = 300

        self.assertEqual(ns0, ns1)

        # add sections
        ns0.sects = [NcsSection(0, 0, 100, 100, 10)]
        ns1.sects = [NcsSection(0, 0, 100, 100, 10)]

        self.assertEqual(ns0, ns1)

        # check inequality for different attributes
        # different number of sections
        ns0.sects.append(NcsSection(0, 0, 100, 100, 10))
        self.assertNotEqual(ns0, ns1)

        # different section attributes
        ns0.sects = [NcsSection(0, 0, 200, 200, 10)]
        self.assertNotEqual(ns0, ns1)

        # different attributes
        ns0.sampFreqUsed = 400
        self.assertNotEqual(ns0, ns1)


# I comment this now and will put it back when files will be in gin.g-node
# class TestNlxHeader(TestNeuralynxRawIO, unittest.TestCase):
#     def test_no_date_time(self):
#         filename = self.get_local_path("neuralynx/NoDateHeader/NoDateHeader.nev")

#         with self.assertRaises(IOError):
#             hdr = NlxHeader(filename)

#         hdr = NlxHeader(filename, props_only=True)

#         self.assertEqual(len(hdr), 11)
#         self.assertEqual(hdr['ApplicationName'], 'Pegasus')
#         self.assertEqual(hdr['FileType'], 'Event')

if __name__ == "__main__":
    unittest.main()

    # test = TestNeuralynxRawIO()
    # test.test_scan_ncs_files()
    # test.test_exclude_filenames()
    # test.test_include_filenames()

    # test = TestNcsSectionsFactory()
    # test.test_ncsblocks_partial()
    # test.test_build_given_actual_frequency()



"""
This module defines multiple utility functions for filtering, creation, slicing,
etc. of neo.core objects.
"""

import copy
import warnings

import numpy as np
import quantities as pq

import neo

reserved_annotations = ["nix_name"]


def get_events(container, **properties):
    """
    This function returns a list of Event objects, corresponding to given
    key-value pairs in the attributes or annotations of the Event.

    Parameters
    ----------
    container: Block or Segment
        The Block or Segment object to extract data from.

    Keyword Arguments:
    ------------------
    The Event properties to filter for.
    Each property name is matched to an attribute or an
    (array-)annotation of the Event. The value of property corresponds
    to a valid entry or a list of valid entries of the attribute or
    (array-)annotation.

    If the value is a list of entries of the same
    length as the number of events in the Event object, the list entries
    are matched to the events in the Event object. The resulting Event
    object contains only those events where the values match up.

    Otherwise, the value is compared to the attribute or (array-)annotation
    of the Event object as such, and depending on the comparison, either the
    complete Event object is returned or not.

    If no keyword arguments is passed, all Event Objects will
    be returned in a list.

    Returns
    -------
    events: list
        A list of Event objects matching the given criteria.

    Examples
    --------
        >>> import neo
        >>> from neo.utils import get_events
        >>> import quantities as pq
        >>> event = neo.Event(times=[0.5, 10.0, 25.2] * pq.s)
        >>> event.annotate(event_type='trial start')
        >>> event.array_annotate(trial_id=[1, 2, 3])
        >>> seg = neo.Segment()
        >>> seg.events = [event]

        # Will return a list with the complete event object
        >>> get_events(seg, event_type='trial start')

        # Will return an empty list
        >>> get_events(seg, event_type='trial stop')

        # Will return a list with an Event object, but only with trial 2
        >>> get_events(seg, trial_id=2)

        # Will return a list with an Event object, but only with trials 1 and 2
        >>> get_events(seg, trial_id=[1, 2])
    """
    if isinstance(container, neo.Segment):
        return _get_from_list(container.events, prop=properties)

    elif isinstance(container, neo.Block):
        event_lst = []
        for seg in container.segments:
            event_lst += _get_from_list(seg.events, prop=properties)
        return event_lst
    else:
        raise TypeError(
            f"Container needs to be of type Block or Segment, not {type(container)} " "in order to extract Events."
        )


def get_epochs(container, **properties):
    """
    This function returns a list of Epoch objects, corresponding to given
    key-value pairs in the attributes or annotations of the Epoch.

    Parameters
    ----------
    container: Block or Segment
        The Block or Segment object to extract data from.

    Keyword Arguments:
    ------------------
    The Epoch properties to filter for.
    Each property name is matched to an attribute or an
    (array-)annotation of the Epoch. The value of property corresponds
    to a valid entry or a list of valid entries of the attribute or
    (array-)annotation.

    If the value is a list of entries of the same
    length as the number of epochs in the Epoch object, the list entries
    are matched to the epochs in the Epoch object. The resulting Epoch
    object contains only those epochs where the values match up.

    Otherwise, the value is compared to the attribute or (array-)annotation
    of the Epoch object as such, and depending on the comparison, either the
    complete Epoch object is returned or not.

    If no keyword arguments is passed, all Epoch Objects will
    be returned in a list.

    Returns
    -------
    epochs: list
        A list of Epoch objects matching the given criteria.

    Examples
    --------
        >>> import neo
        >>> from neo.utils import get_epochs
        >>> import quantities as pq
        >>> epoch = neo.Epoch(times=[0.5, 10.0, 25.2] * pq.s,
        ...                   durations=[100, 100, 100] * pq.ms,
        ...                   epoch_type='complete trial')
        >>> epoch.array_annotate(trial_id=[1, 2, 3])
        >>> seg = neo.Segment()
        >>> seg.epochs = [epoch]

        # Will return a list with the complete event object
        >>> get_epochs(seg, epoch_type='complete trial')

        # Will return an empty list
        >>> get_epochs(seg, epoch_type='error trial')

        # Will return a list with an Event object, but only with trial 2
        >>> get_epochs(seg, trial_id=2)

        # Will return a list with an Event object, but only with trials 1 and 2
        >>> get_epochs(seg, trial_id=[1, 2])
    """
    if isinstance(container, neo.Segment):
        return _get_from_list(container.epochs, prop=properties)

    elif isinstance(container, neo.Block):
        epoch_list = []
        for seg in container.segments:
            epoch_list += _get_from_list(seg.epochs, prop=properties)
        return epoch_list
    else:
        raise TypeError(
            f"Container needs to be of type Block or Segment, not {type(container)} " "in order to extract Epochs."
        )


def _get_from_list(input_list, prop=None):
    """
    Internal function
    """
    output_list = []
    # empty or no dictionary
    if not prop or bool([b for b in prop.values() if b == []]):
        output_list += [e for e in input_list]
    # dictionary is given
    else:
        for ep in input_list:
            if isinstance(ep, neo.Epoch) or isinstance(ep, neo.Event):
                sparse_ep = ep.copy()
            elif isinstance(ep, neo.io.proxyobjects.EpochProxy) or isinstance(ep, neo.io.proxyobjects.EventProxy):
                # need to load the Event/Epoch in order to be able to filter by array annotations
                sparse_ep = ep.load()
            for k in prop.keys():
                sparse_ep = _filter_event_epoch(sparse_ep, k, prop[k])
                # if there is nothing left, it cannot filtered
                if sparse_ep is None:
                    break
            if sparse_ep is not None:
                output_list.append(sparse_ep)
    return output_list


def _filter_event_epoch(obj, annotation_key, annotation_value):
    """
    Internal function.

    This function returns a copy of a Event or Epoch object, which only
    contains attributes or annotations corresponding to requested key-value
    pairs.

    Parameters
    ----------
    obj : Event
        The Event or Epoch object to modify.
    annotation_key : string, int or float
        The name of the annotation used to filter.
    annotation_value : string, int, float, list or np.ndarray
        The accepted value or list of accepted values of the attributes or
        annotations specified by annotation_key. For each entry in obj the
        respective annotation defined by annotation_key is compared to the
        annotation value. The entry of obj is kept if the attribute or
        annotation is equal or contained in annotation_value.

    Returns
    -------
    obj : Event or Epoch
        The Event or Epoch object with every event or epoch removed that does
        not match the filter criteria (i.e., where none of the entries in
        annotation_value match the attribute or annotation annotation_key.
    """
    valid_ids = _get_valid_ids(obj, annotation_key, annotation_value)

    if len(valid_ids) == 0:
        return None

    return _event_epoch_slice_by_valid_ids(obj, valid_ids)


def _event_epoch_slice_by_valid_ids(obj, valid_ids):
    """
    Internal function
    """
    if type(obj) is neo.Event or type(obj) is neo.Epoch:
        sparse_obj = copy.deepcopy(obj[valid_ids])
    else:
        raise TypeError("Can only slice Event and Epoch objects by valid IDs.")

    return sparse_obj


def _get_valid_ids(obj, annotation_key, annotation_value):
    """
    Internal function
    """

    valid_mask = np.zeros(obj.shape)

    if annotation_key in obj.annotations and obj.annotations[annotation_key] == annotation_value:
        valid_mask = np.ones(obj.shape)

    elif annotation_key == "labels":
        # wrap annotation value to be list
        if not type(annotation_value) in [list, np.ndarray]:
            annotation_value = [annotation_value]
        valid_mask = np.isin(obj.labels, annotation_value)

    elif annotation_key in obj.array_annotations:
        # wrap annotation value to be list
        if not type(annotation_value) in [list, np.ndarray]:
            annotation_value = [annotation_value]
        valid_mask = np.isin(obj.array_annotations[annotation_key], annotation_value)

    elif hasattr(obj, annotation_key) and getattr(obj, annotation_key) == annotation_value:
        valid_mask = np.ones(obj.shape)

    valid_ids = np.where(valid_mask)[0]

    return valid_ids


def add_epoch(segment, event1, event2=None, pre=0 * pq.s, post=0 * pq.s, attach_result=True, **kwargs):
    """
    Create Epochs around a single Event, or between pairs of events. Starting
    and end time of the Epoch can be modified using pre and post as offsets
    before the and after the event(s). Additional keywords will be directly
    forwarded to the Epoch intialization.

    Parameters
    ----------
    segment : Segment
        The segment in which the final Epoch object is added.
    event1 : Event
        The Event objects containing the start events of the epochs. If no
        event2 is specified, these event1 also specifies the stop events, i.e.,
        the Epoch is cut around event1 times.
    event2: Event
        The Event objects containing the stop events of the epochs. If no
        event2 is specified, event1 specifies the stop events, i.e., the Epoch
        is cut around event1 times. The number of events in event2 must match
        that of event1.
    pre, post: Quantity (time)
        Time offsets to modify the start (pre) and end (post) of the resulting
        Epoch. Example: pre=-10*ms and post=+25*ms will cut from 10 ms before
        event1 times to 25 ms after event2 times
    attach_result: bool
        If True, the resulting Epoch object is added to segment.
    **kwargs
        Additional keyword arguments passed to the Epoch object.

    Returns
    -------
    epoch: Epoch
        An Epoch object with the calculated epochs (one per entry in event1).

    See also
    --------
    Event.to_epoch()
    """
    if event2 is None:
        event2 = event1

    if not isinstance(segment, neo.Segment):
        raise TypeError(f"Segment has to be of type Segment, not {type(segment)}")

    # load the full event if a proxy object has been given as an argument
    if isinstance(event1, neo.io.proxyobjects.EventProxy):
        event1 = event1.load()
    if isinstance(event2, neo.io.proxyobjects.EventProxy):
        event2 = event2.load()

    for event in [event1, event2]:
        if not isinstance(event, neo.Event):
            raise TypeError(f"Events have to be of type Event, not {type(event)}")

    if len(event1) != len(event2):
        raise ValueError(
            "event1 and event2 have to have the same number of entries in "
            "order to create epochs between pairs of entries. Match your "
            "events before generating epochs. Current event lengths "
            f"are {len(event1)} and {len(event2)}"
        )

    times = event1.times + pre
    durations = event2.times + post - times

    if any(durations < 0):
        raise ValueError(f"Can not create epoch with negative duration. " "Requested durations {durations}.")
    elif any(durations == 0):
        raise ValueError("Can not create epoch with zero duration.")

    if "name" not in kwargs:
        kwargs["name"] = "epoch"
    if "labels" not in kwargs:
        kwargs["labels"] = [f"{kwargs['name']}_{i}" for i in range(len(times))]

    ep = neo.Epoch(times=times, durations=durations, **kwargs)

    ep.annotate(**clean_annotations(event1.annotations))
    ep.array_annotate(**clean_annotations(event1.array_annotations))

    if attach_result:
        segment.epochs.append(ep)
        segment.check_relationships()

    return ep


def match_events(event1, event2):
    """
    Finds pairs of Event entries in event1 and event2 with the minimum delay,
    such that the entry of event1 directly precedes the entry of event2.
    Returns filtered two events of identical length, which contain matched
    entries.

    Parameters
    ----------
    event1, event2: Event
        The two Event objects to match up.

    Returns
    -------
    event1, event2: Event
        Event objects with identical number of events, containing only those
        events that could be matched against each other. A warning is issued if
        not all events in event1 or event2 could be matched.
    """
    # load the full event if a proxy object has been given as an argument
    if isinstance(event1, neo.io.proxyobjects.EventProxy):
        event1 = event1.load()
    if isinstance(event2, neo.io.proxyobjects.EventProxy):
        event2 = event2.load()

    id1, id2 = 0, 0
    match_ev1, match_ev2 = [], []
    while id1 < len(event1) and id2 < len(event2):
        time1 = event1.times[id1]
        time2 = event2.times[id2]

        # wrong order of events
        if time1 >= time2:
            id2 += 1

        # shorter epoch possible by later event1 entry
        elif id1 + 1 < len(event1) and event1.times[id1 + 1] < time2:
            # there is no event in 2 until the next event in 1
            id1 += 1

        # found a match
        else:
            match_ev1.append(id1)
            match_ev2.append(id2)
            id1 += 1
            id2 += 1

    if id1 < len(event1):
        warnings.warn(
            "Could not match all events to generate epochs. Missed " f"{len(event1) - id1} event entries in event1 list"
        )
    if id2 < len(event2):
        warnings.warn(
            "Could not match all events to generate epochs. Missed " f"{len(event2) - id2} event entries in event2 list"
        )

    event1_matched = _event_epoch_slice_by_valid_ids(obj=event1, valid_ids=match_ev1)
    event2_matched = _event_epoch_slice_by_valid_ids(obj=event2, valid_ids=match_ev2)

    return event1_matched, event2_matched


def cut_block_by_epochs(block, properties=None, reset_time=False):
    """
    This function cuts Segments in a Block according to multiple Neo
    Epoch objects.

    The function alters the Block by adding one Segment per Epoch entry
    fulfilling a set of conditions on the Epoch attributes and annotations. The
    original segments are removed from the block.

    A dictionary contains restrictions on which Epochs are considered for
    the cutting procedure. To this end, it is possible to
    specify accepted (valid) values of specific annotations on the source
    Epochs.

    The resulting cut segments may either retain their original time stamps, or
    be shifted to a common starting time.

    Parameters
    ----------
    block: Block
        Contains the Segments to cut according to the Epoch criteria provided
    properties: dictionary
        A dictionary that contains the Epoch keys and values to filter for.
        Each key of the dictionary is matched to an attribute or an
        annotation or an array_annotation of the Event.
        The value of each dictionary entry corresponds to a valid entry or a
        list of valid entries of the attribute or (array) annotation.

        If the value belonging to the key is a list of entries of the same
        length as the number of epochs in the Epoch object, the list entries
        are matched to the epochs in the Epoch object. The resulting Epoch
        object contains only those epochs where the values match up.

        Otherwise, the value is compared to the attributes or annotation of the
        Epoch object as such, and depending on the comparison, either the
        complete Epoch object is returned or not.

        If None or an empty dictionary is passed, all Epoch Objects will
        be considered

    reset_time: bool
        If True the times stamps of all sliced objects are set to fall
        in the range from 0 to the duration of the epoch duration.
        If False, original time stamps are retained.
        Default is False.

    Returns
    -------
    new_block : Block
        Updated block.
    """
    if not isinstance(block, neo.Block):
        raise TypeError(f"block needs to be a Block, not {type(block)}")

    new_block = neo.Block()

    for seg in block.segments:
        epochs = _get_from_list(seg.epochs, prop=properties)
        if len(epochs) > 1:
            warnings.warn(
                f"Segment {seg.name} contains multiple epochs with "
                f"requested properties ({properties}). Sub-segments can "
                "have overlapping times"
            )

        elif len(epochs) == 0:
            warnings.warn(
                f"No epoch is matching the requested epoch properties {properties}. "
                f"No cutting of segment {seg.name} performed."
            )

        for epoch in epochs:
            new_segments = cut_segment_by_epoch(seg, epoch=epoch, reset_time=reset_time)
            new_block.segments.extend(new_segments)

    new_block.check_relationships()

    return new_block


def cut_segment_by_epoch(seg, epoch, reset_time=False):
    """
    Cuts a Segment according to an Epoch object

    The function returns a list of Segments, where each segment corresponds
    to an epoch in the Epoch object and contains the data of the original
    Segment cut to that particular Epoch.

    The resulting segments may either retain their original time stamps,
    or can be shifted to a common time axis.

    Parameters
    ----------
    seg: Segment
        The Segment containing the original uncut data.
    epoch: Epoch
        For each epoch in this input, one segment is generated according to
         the epoch time and duration.
    reset_time: bool
        If True the times stamps of all sliced objects are set to fall
        in the range from 0 to the duration of the epoch duration.
        If False, original time stamps are retained.
        Default is False.

    Returns
    -------
    segments: list of Segments
        Per epoch in the input, a Segment with AnalogSignal and/or
        SpikeTrain Objects will be generated and returned. Each Segment will
        receive the annotations of the corresponding epoch in the input.
    """
    if not isinstance(seg, neo.Segment):
        raise TypeError(f"Seg needs to be of type Segment, not {type(seg)}")

    if not isinstance(epoch, neo.Epoch):
        raise TypeError(f"Epoch needs to be of type Epoch, not {type(epoch)}")

    segments = []
    for ep_id in range(len(epoch)):
        subseg = seg.time_slice(epoch.times[ep_id], epoch.times[ep_id] + epoch.durations[ep_id], reset_time=reset_time)

        subseg.annotations = clean_annotations(subseg.annotations)
        subseg.annotate(**clean_annotations(epoch.annotations))

        # Add array-annotations of Epoch
        for key, val in clean_annotations(epoch.array_annotations).items():
            if len(val):
                subseg.annotations[key] = copy.copy(val[ep_id])

        segments.append(subseg)

    return segments


def clean_annotations(dictionary):
    """
    Remove reserved keys from an annotation dictionary.

    Parameters
    ----------
    dictionary: dict
        annotation dictionary to be cleaned

    Returns
    -------
    dict
        A cleaned version of the annotations
    """
    return {k: v for k, v in dictionary.items() if k not in reserved_annotations}


def is_block_rawio_compatible(block, return_problems=False):
    """
    The neo.rawio layer have some restriction compared to neo.io layer:
      * consistent channels across segments
      * no IrregularlySampledSignal
      * consistent sampling rate across segments

    This function tests if a neo.Block that could be written in a nix file could be read
    back with the NIXRawIO.

    Parameters
    ----------
    block: Block
        A block
    return_problems: bool (False by default)
        Controls whether a list of str that describe problems is also provided as return value

    Returns
    -------
    is_rawio_compatible: bool
        Compatible or not.
    problems: list of str
        Optional, depending on value of `return_problems`.
        A list that describe problems for rawio compatibility.
    """
    if len(block.segments) == 0:
        raise neo.core.NeoReadWriteError("This block doesn't have segments")

    problems = []

    # check that all Segments have the same number of object.
    n_sig = len(block.segments[0].analogsignals)
    n_st = len(block.segments[0].spiketrains)
    n_ev = len(block.segments[0].events)
    n_ep = len(block.segments[0].epochs)
    sig_count_consistent = True
    for seg in block.segments:
        if len(seg.analogsignals) != n_sig:
            problems.append("Number of AnalogSignals is not consistent across segments")
            sig_count_consistent = False
        if len(seg.spiketrains) != n_st:
            problems.append("Number of SpikeTrains is not consistent across segments")
        if len(seg.events) != n_ev:
            problems.append("Number of Events is not consistent across segments")
        if len(seg.epochs) != n_ep:
            problems.append("Number of Epochs is not consistent across segments")

    # check for AnalogSigal that sampling_rate/units/number of channel
    # is consistent across segments.
    if sig_count_consistent:
        seg0 = block.segments[0]
        for i in range(n_sig):
            for seg in block.segments:
                if seg.analogsignals[i].sampling_rate != seg0.analogsignals[i].sampling_rate:
                    problems.append("AnalogSignals have inconsistent sampling rate across segments")
                if seg.analogsignals[i].shape[1] != seg0.analogsignals[i].shape[1]:
                    problems.append("AnalogSignals have inconsistent channel count across segments")
                if seg.analogsignals[i].units != seg0.analogsignals[i].units:
                    problems.append("AnalogSignals have inconsistent units across segments")

    # check no IrregularlySampledSignal
    for seg in block.segments:
        if len(seg.irregularlysampledsignals) > 0:
            problems.append("IrregularlySampledSignals are not raw compatible")

    # returns
    is_rawio_compatible = len(problems) == 0
    if return_problems:
        return is_rawio_compatible, problems
    else:
        return is_rawio_compatible



"""
This module implements the ObjectList class, which is used to peform type checks
and handle relationships within the Neo Block-Segment-Data hierarchy.
"""

import sys

from neo.core.baseneo import BaseNeo


class ObjectList:
    """
    This class behaves like a list, but has additional functionality
    to handle relationships within Neo hierarchy, and perform type checks.
    """

    def __init__(self, allowed_contents, parent=None):
        # validate allowed_contents and normalize it to a tuple
        if isinstance(allowed_contents, type) and issubclass(allowed_contents, BaseNeo):
            self.allowed_contents = (allowed_contents,)
        else:
            for item in allowed_contents:
                if not issubclass(item, BaseNeo):
                    raise TypeError("Each item in allowed_contents must be a subclass of BaseNeo")
            self.allowed_contents = tuple(allowed_contents)
        self._items = []
        self.parent = parent

    def _handle_append(self, obj):
        if not (
            isinstance(obj, self.allowed_contents)
            or (  # also allow proxy objects of the correct type
                hasattr(obj, "proxy_for") and obj.proxy_for in self.allowed_contents
            )
        ):
            raise TypeError(f"Object is a {type(obj)}. It should be one of {self.allowed_contents}.")

        if self._contains(obj):
            raise ValueError("Cannot add this object because it is already contained within the list")

        # set the child-parent relationship
        if self.parent:
            relationship_name = self.parent.__class__.__name__.lower()
            if relationship_name == "group":
                raise Exception("Objects in groups should not link to the group as their parent")
            current_parent = getattr(obj, relationship_name)
            if current_parent != self.parent:
                # use weakref here? - see https://github.com/NeuralEnsemble/python-neo/issues/684
                setattr(obj, relationship_name, self.parent)

    def _contains(self, obj):
        if self._items is None:
            obj_ids = []
        else:
            obj_ids = [id(item) for item in self._items]
        return id(obj) in obj_ids

    def __str__(self):
        return str(self._items)

    def __repr__(self):
        return repr(self._items)

    def __add__(self, objects):
        # todo: decision: return a list, or a new DataObjectList?
        if isinstance(objects, ObjectList):
            return self._items + objects._items
        else:
            return self._items + objects

    def __radd__(self, objects):
        if isinstance(objects, ObjectList):
            return objects._items + self._items
        else:
            return objects + self._items

    def __contains__(self, key):
        return key in self._items

    def __iadd__(self, objects):
        for obj in objects:
            self._handle_append(obj)
        self._items.extend(objects)
        return self

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, i):
        return self._items[i]

    def __len__(self):
        return len(self._items)

    def __setitem__(self, key, value):
        self._items[key] = value

    def append(self, obj):
        self._handle_append(obj)
        self._items.append(obj)

    def extend(self, objects):
        for obj in objects:
            self._handle_append(obj)
        self._items.extend(objects)

    def clear(self):
        self._items = []

    def count(self, value):
        return self._items.count(value)

    def index(self, value, start=0, stop=sys.maxsize):
        return self._items.index(value, start, stop)

    def insert(self, index, obj):
        self._handle_append(obj)
        self._items.insert(index, obj)

    def pop(self, index=-1):
        return self._items.pop(index)

    def remove(self, value):
        return self._items.remove(value)

    def reverse(self):
        raise self._items.reverse()

    def sort(self, *args, key=None, reverse=False):
        self._items.sort(*args, key=key, reverse=reverse)



"""
Tests of neo.rawio.mearecrawio

"""

import unittest

from neo.rawio.plexon2rawio import Plexon2RawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


try:
    from neo.rawio.plexon2rawio.pypl2 import pypl2lib

    HAVE_PYPL2 = True
except (ImportError, TimeoutError):
    HAVE_PYPL2 = False


@unittest.skipUnless(HAVE_PYPL2, "requires pypl package and all its dependencies")
class TestPlexon2RawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = Plexon2RawIO
    entities_to_download = ["plexon"]
    entities_to_test = ["plexon/4chDemoPL2.pl2"]


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
from tempfile import TemporaryDirectory
import platform
from neo.io import get_io, list_candidate_ios, NixIO
import pytest

try:
    import nixio

    HAVE_NIX = True
except:
    HAVE_NIX = False


def test_list_candidate_ios_non_existant_file():
    # use plexon io suffix for testing here
    non_existant_file = Path("non_existant_folder/non_existant_file.plx")
    non_existant_file.unlink(missing_ok=True)
    ios = list_candidate_ios(non_existant_file)

    assert ios

    # cleanup
    non_existant_file.unlink(missing_ok=True)


def test_list_candidate_ios_filename_stub():
    # create dummy folder with dummy files
    with TemporaryDirectory(prefix="filename_stub_test_") as test_folder:
        test_folder = Path(test_folder)
        test_filename = test_folder / "dummy_file.nix"
        test_filename.touch()
        filename_stub = test_filename.with_suffix("")

        # check that io is found even though file suffix was not provided
        ios = list_candidate_ios(filename_stub)

        assert NixIO in ios


@pytest.mark.skipif(not HAVE_NIX or platform.system() == "Windows", reason="Need nixio in order to return NixIO class")
def test_get_io_non_existant_file_writable_io():
    # use nixio for testing with writable io
    non_existant_file = Path("non_existant_file.nix")
    non_existant_file.unlink(missing_ok=True)
    io = get_io(non_existant_file)

    assert isinstance(io, NixIO)

    # cleanup
    non_existant_file.unlink(missing_ok=True)  # cleanup will fail on Windows so need to skip



"""
Class for reading/writing data in a raw binary interleaved compact file.
Sampling rate, units, number of channel and dtype must be externally known.
This generic format is quite widely used in old acquisition systems and is quite universal
for sharing data.

Supported : Read/Write

Author: sgarcia

"""

import numpy as np

from neo.core import Segment, AnalogSignal, NeoReadWriteError

from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.rawbinarysignalrawio import RawBinarySignalRawIO


class RawBinarySignalIO(RawBinarySignalRawIO, BaseFromRaw):
    """
    Class for reading/writing data in a raw binary interleaved compact file.

    **Important release note**

    Since the version neo 0.6.0 and the neo.rawio API,
    argmuents of the IO (dtype, nb_channel, sampling_rate) must be
    given at the __init__ and not at read_segment() because there is
    no read_segment() in neo.rawio classes.

    So now the usage is:
        >>>>r = io.RawBinarySignalIO(filename='file.raw', dtype='int16',
                                    nb_channel=16, sampling_rate=10000.)


    """

    _prefered_signal_group_mode = "group-by-same-units"

    is_readable = True
    is_writable = True

    supported_objects = [Segment, AnalogSignal]
    readable_objects = [Segment]
    writeable_objects = [Segment]

    def __init__(
        self,
        filename,
        dtype="int16",
        sampling_rate=10000.0,
        nb_channel=2,
        signal_gain=1.0,
        signal_offset=0.0,
        bytesoffset=0,
    ):
        RawBinarySignalRawIO.__init__(
            self,
            filename=filename,
            dtype=dtype,
            sampling_rate=sampling_rate,
            nb_channel=nb_channel,
            signal_gain=signal_gain,
            signal_offset=signal_offset,
            bytesoffset=bytesoffset,
        )
        BaseFromRaw.__init__(self, filename)

    def write_segment(self, segment):
        """

        **Arguments**
            segment : the segment to write. Only analog signals will be written.

        Support only 2 cases:
          * segment.analogsignals have one 2D AnalogSignal
          * segment.analogsignals have several 1D AnalogSignal with
            same length/sampling_rate/dtype

        """

        if self.bytesoffset:
            raise NotImplementedError("bytesoffset values other than 0 " + "not supported")

        anasigs = segment.analogsignals
        if len(anasigs) == 0:
            raise NeoReadWriteError("No AnalogSignal to write")

        anasig0 = anasigs[0]
        if len(anasigs) == 1 and anasig0.ndim == 2:
            numpy_sigs = anasig0.magnitude
        else:

            if anasig0.ndim != 1 and (anasig0.ndim == 2 and anasig0.shape[1] != 1):
                raise NeoReadWriteError("Incorrect analogsignal shaping for write")
            # all AnaologSignal from Segment must have the same length/sampling_rate/dtype
            for anasig in anasigs[1:]:
                if anasig.shape != anasig0.shape:
                    raise ValueError("The shape of one of the analog signals does not match")
                if anasig.sampling_rate != anasig0.sampling_rate:
                    raise ValueError("The sampling_rate of one of the analog signals does not match")
                if anasig.dtype != anasig0.dtype:
                    raise ValueError("The dtype of one the analog signals does not match")

            numpy_sigs = np.empty((anasig0.size, len(anasigs)))
            for i, anasig in enumerate(anasigs):
                numpy_sigs[:, i] = anasig.magnitude.flatten()

        numpy_sigs -= self.signal_offset
        numpy_sigs /= self.signal_gain
        numpy_sigs = numpy_sigs.astype(self.dtype)

        with open(self.filename, "wb") as f:
            f.write(numpy_sigs.tobytes())



"""
Tests of the neo.core.epoch.Epoch class
"""

import unittest
import warnings
from copy import deepcopy

import numpy as np
import quantities as pq
import pickle
import os
from numpy.testing import assert_array_equal

from neo.core.dataobject import ArrayDict

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

from neo.core.epoch import Epoch
from neo.core import Segment
from neo.test.tools import (
    assert_neo_object_is_compliant,
    assert_arrays_equal,
    assert_arrays_almost_equal,
    assert_same_sub_schema,
    assert_same_attributes,
    assert_same_annotations,
    assert_same_array_annotations,
)


class TestEpoch(unittest.TestCase):

    def setUp(self):
        self.params = {"test0": "y1", "test1": ["deeptest"], "test2": True}
        self.seg = Segment()
        self.epc = Epoch(
            times=[10, 20, 30, 40, 50] * pq.s,
            durations=[10, 5, 7, 14, 9] * pq.ms,
            labels=np.array(["btn0", "btn1", "btn2", "btn0", "btn3"], dtype="S"),
            **self.params,
        )
        self.epc.segment = self.seg

    def test_Epoch_creation(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"names": ["a", "b", "c"], "index": np.arange(10, 13)}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        assert_arrays_equal(epc.times, [1.1, 1.5, 1.7] * pq.ms)
        assert_arrays_equal(epc.durations, [20, 40, 60] * pq.ns)
        assert_arrays_equal(epc.labels, np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"))
        self.assertEqual(epc.name, "test")
        self.assertEqual(epc.description, "tester")
        self.assertEqual(epc.file_origin, "test.file")
        self.assertEqual(epc.annotations["test0"], [1, 2])
        self.assertEqual(epc.annotations["test1"], 1.1)
        self.assertEqual(epc.annotations["test2"], "y1")
        self.assertTrue(epc.annotations["test3"])
        assert_arrays_equal(epc.array_annotations["names"], np.array(["a", "b", "c"]))
        assert_arrays_equal(epc.array_annotations["index"], np.arange(10, 13))
        self.assertIsInstance(epc.array_annotations, ArrayDict)

    def test_Epoch_invalid_times_dimension(self):
        data2d = np.array([1, 2, 3, 4]).reshape((4, -1))
        durations = np.array([1, 1, 1, 1])
        self.assertRaises(ValueError, Epoch, times=data2d * pq.s, durations=durations)

    def test_Epoch_creation_invalid_durations_labels(self):
        self.assertRaises(ValueError, Epoch, [1.1, 1.5, 1.7] * pq.ms, durations=[20, 40, 60, 80] * pq.ns)
        self.assertRaises(ValueError, Epoch, [1.1, 1.5, 1.7] * pq.ms, durations=[20, 40, 60] * pq.ns, labels=["A", "B"])

    def test_Epoch_creation_scalar_duration(self):
        # test with scalar for durations
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=20 * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
        )
        assert_neo_object_is_compliant(epc)

        assert_arrays_equal(epc.times, [1.1, 1.5, 1.7] * pq.ms)
        assert_arrays_equal(epc.durations, [20, 20, 20] * pq.ns)
        self.assertEqual(epc.durations.size, 3)
        assert_arrays_equal(epc.labels, np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"))

    def test_Epoch_creation_from_lists(self):
        epc = Epoch([1.1, 1.5, 1.7], [20.0, 20.0, 20.0], ["test event 1", "test event 2", "test event 3"], units=pq.ms)
        assert_arrays_equal(epc.times, [1.1, 1.5, 1.7] * pq.ms)
        assert_arrays_equal(epc.durations, [20.0, 20.0, 20.0] * pq.ms)
        assert_arrays_equal(epc.labels, np.array(["test event 1", "test event 2", "test event 3"]))

    def test_Epoch_repr(self):
        params = {"test2": "y1", "test3": True}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = (
            "<Epoch: test epoch 1@1.1 ms for 20.0 ns, "
            + "test epoch 2@1.5 ms for 40.0 ns, "
            + "test epoch 3@1.7 ms for 60.0 ns>"
        )

        res = repr(epc)

        self.assertEqual(targ, res)

    def test_Epoch_merge(self):
        params1 = {"test2": "y1", "test3": True}
        params2 = {"test2": "no", "test4": False}
        paramstarg = {"test2": "yes;no", "test3": True, "test4": False}
        arr_ann1 = {"index": np.arange(10, 13)}
        arr_ann2 = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc1 = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.us,
            labels=np.array(["test epoch 1 1", "test epoch 1 2", "test epoch 1 3"], dtype="U"),
            name="test",
            description="tester 1",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann1,
            **params1,
        )
        epc2 = Epoch(
            [2.1, 2.5, 2.7] * pq.us,
            durations=[3, 5, 7] * pq.ms,
            labels=np.array(["test epoch 2 1", "test epoch 2 2", "test epoch 2 3"], dtype="U"),
            name="test",
            description="tester 2",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann2,
            **params2,
        )
        epctarg = Epoch(
            [1.1, 1.5, 1.7, 0.0021, 0.0025, 0.0027] * pq.ms,
            durations=[20, 40, 60, 3000, 5000, 7000] * pq.us,
            labels=np.array(
                [
                    "test epoch 1 1",
                    "test epoch 1 2",
                    "test epoch 1 3",
                    "test epoch 2 1",
                    "test epoch 2 2",
                    "test epoch 2 3",
                ],
                dtype="U",
            ),
            name="test",
            description="merge(tester 1, tester 2)",
            file_origin="test.file",
            array_annotations={"index": [10, 11, 12, 0, 1, 2]},
            test1=1,
            **paramstarg,
        )
        assert_neo_object_is_compliant(epc1)
        assert_neo_object_is_compliant(epc2)
        assert_neo_object_is_compliant(epctarg)

        with warnings.catch_warnings(record=True) as w:
            epcres = epc1.merge(epc2)

            self.assertTrue(len(w), 1)
            self.assertEqual(w[0].category, UserWarning)
            self.assertSequenceEqual(
                str(w[0].message),
                "The following array annotations were "
                "omitted, because they were only present"
                " in one of the merged objects: "
                "[] from the one that was merged "
                "into and ['test'] from the one that "
                "was merged into the other",
            )

        assert_neo_object_is_compliant(epcres)
        assert_same_sub_schema(epctarg, epcres)
        # Remove this, when array_annotations are added to assert_same_sub_schema
        assert_arrays_equal(epcres.array_annotations["index"], np.array([10, 11, 12, 0, 1, 2]))
        self.assertTrue("test" not in epcres.array_annotations)
        self.assertIsInstance(epcres.array_annotations, ArrayDict)

    def test_set_labels_duration(self):
        epc = Epoch([1.1, 1.5, 1.7] * pq.ms, durations=20 * pq.ns, labels=["A", "B", "C"])
        assert_array_equal(epc.durations.magnitude, np.array([20, 20, 20]))
        epc.durations = [20.0, 21.0, 22.0] * pq.ns
        assert_array_equal(epc.durations.magnitude, np.array([20, 21, 22]))
        self.assertRaises(ValueError, setattr, epc, "durations", [25.0, 26.0] * pq.ns)

        assert_array_equal(epc.labels, np.array(["A", "B", "C"]))
        epc.labels = ["D", "E", "F"]
        assert_array_equal(epc.labels, np.array(["D", "E", "F"]))
        self.assertRaises(ValueError, setattr, epc, "labels", ["X", "Y"])

    def test__children(self):
        params = {"test2": "y1", "test3": True}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        segment = Segment(name="seg1")
        segment.epochs = [epc]
        segment.check_relationships()

        self.assertEqual(epc._parent_objects, ("Segment",))

        self.assertEqual(epc._parent_containers, ("segment",))

        self.assertEqual(epc._parent_objects, ("Segment",))
        self.assertEqual(epc._parent_containers, ("segment",))

        self.assertEqual(len(epc.parents), 1)
        self.assertEqual(epc.parents[0].name, "seg1")

        assert_neo_object_is_compliant(epc)

    @unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
    def test__pretty(self):
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        prepr = pretty(epc)
        targ = (
            f"Epoch containing {len(epc)} epochs with labels; "
            f"time units {epc.units.dimensionality.string}; datatype float64 "
            f"\nname: '{epc.name}'\ndescription: '{epc.description}'"
            f"\nannotations: {pretty(epc.annotations)}"
        )

        self.assertEqual(prepr, targ)

    def test__time_slice(self):
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            times=[10, 20, 30] * pq.s,
            durations=[10, 5, 7] * pq.ms,
            labels=np.array(["btn0", "btn1", "btn2"], dtype="U"),
            foo="bar",
            array_annotations=arr_ann,
        )

        epc2 = epc.time_slice(10 * pq.s, 20 * pq.s)
        assert_arrays_equal(epc2.times, [10, 20] * pq.s)
        assert_arrays_equal(epc2.durations, [10, 5] * pq.ms)
        assert_arrays_equal(epc2.labels, np.array(["btn0", "btn1"], dtype="U"))
        self.assertEqual(epc.annotations, epc2.annotations)
        assert_arrays_equal(epc2.array_annotations["index"], np.arange(2))
        assert_arrays_equal(epc2.array_annotations["test"], np.array(["a", "b"]))
        self.assertIsInstance(epc2.array_annotations, ArrayDict)

    def test_time_slice2(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [1.5] * pq.ms,
            durations=[40] * pq.ns,
            labels=np.array(["test epoch 2"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations={"index": [1], "test": ["b"]},
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = 1.2
        t_stop = 1.6
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.array([1]))
        assert_arrays_equal(result.array_annotations["test"], np.array(["b"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test__time_slice_deepcopy_annotations(self):
        params1 = {"test0": "y1", "test1": ["deeptest"], "test2": True}
        self.epc.annotate(**params1)
        # time_slice spike train, keep sliced spike times
        t_start = 15 * pq.s
        t_stop = 45 * pq.s
        result = self.epc.time_slice(t_start, t_stop)

        # Change annotations of original
        params2 = {"test0": "y2", "test2": False}
        self.epc.annotate(**params2)
        self.epc.annotations["test1"][0] = "shallowtest"

        self.assertNotEqual(self.epc.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(self.epc.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(self.epc.annotations["test2"], result.annotations["test2"])

        # Change annotations of result
        params3 = {"test0": "y3"}
        result.annotate(**params3)
        result.annotations["test1"][0] = "shallowtest2"

        self.assertNotEqual(self.epc.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(self.epc.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(self.epc.annotations["test2"], result.annotations["test2"])

    def test__time_slice_deepcopy_array_annotations(self):
        length = self.epc.shape[-1]
        params1 = {
            "test0": [f"y{i}" for i in range(length)],
            "test1": ["deeptest" for i in range(length)],
            "test2": [(-1) ** i > 0 for i in range(length)],
        }
        self.epc.array_annotate(**params1)
        # time_slice spike train, keep sliced spike times
        t_start = 15 * pq.s
        t_stop = 45 * pq.s
        result = self.epc.time_slice(t_start, t_stop)

        # Change annotations of original
        params2 = {
            "test0": [f"x{i}" for i in range(length)],
            "test2": [(-1) ** (i + 1) > 0 for i in range(length)],
        }
        self.epc.array_annotate(**params2)
        self.epc.array_annotations["test1"][2] = "shallowtest"

        self.assertFalse(all(self.epc.array_annotations["test0"][1:4] == result.array_annotations["test0"]))
        self.assertFalse(all(self.epc.array_annotations["test1"][1:4] == result.array_annotations["test1"]))
        self.assertFalse(all(self.epc.array_annotations["test2"][1:4] == result.array_annotations["test2"]))

        # Change annotations of result
        params3 = {"test0": [f"z{i}" for i in range(1, 4)]}
        result.array_annotate(**params3)
        result.array_annotations["test1"][1] = "shallow2"

        self.assertFalse(all(self.epc.array_annotations["test0"][1:4] == result.array_annotations["test0"]))
        self.assertFalse(all(self.epc.array_annotations["test1"][1:4] == result.array_annotations["test1"]))
        self.assertFalse(all(self.epc.array_annotations["test2"][1:4] == result.array_annotations["test2"]))

    def test__time_slice_deepcopy_data(self):
        result = self.epc.time_slice(None, None)

        # Change values of original array
        self.epc[2] = 7.3 * self.epc.units

        self.assertFalse(all(self.epc == result))

        # Change values of sliced array
        result[3] = 9.5 * result.units

        self.assertFalse(all(self.epc == result))

    def test_time_slice_out_of_boundries(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = 0.0001
        t_stop = 30
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.array(arr_ann["index"]))
        assert_arrays_equal(result.array_annotations["test"], np.array(arr_ann["test"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_empty(self):
        params = {"test2": "y1", "test3": True}
        epc = Epoch(
            [] * pq.ms,
            durations=[] * pq.ns,
            labels=np.array([], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [] * pq.ms,
            durations=[] * pq.ns,
            labels=np.array([], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = 1.2
        t_stop = 1.6
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_stop(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [1.5, 1.7] * pq.ms,
            durations=[40, 60] * pq.ns,
            labels=np.array(["test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations={"index": [1, 2], "test": ["b", "c"]},
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = 1.2
        t_stop = None
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.array([1, 2]))
        assert_arrays_equal(result.array_annotations["test"], np.array(["b", "c"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_start(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [1.1, 1.5] * pq.ms,
            durations=[20, 40] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations={"index": [0, 1], "test": ["a", "b"]},
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = None
        t_stop = 1.6
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.array([0, 1]))
        assert_arrays_equal(result.array_annotations["test"], np.array(["a", "b"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_both(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = None
        t_stop = None
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.array([0, 1, 2]))
        assert_arrays_equal(result.array_annotations["test"], np.array(["a", "b", "c"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_differnt_units(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epc = Epoch(
            [1.1, 1.5, 1.7] * pq.ms,
            durations=[20, 40, 60] * pq.ns,
            labels=np.array(["test epoch 1", "test epoch 2", "test epoch 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        epc.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(epc)

        targ = Epoch(
            [1.5] * pq.ms,
            durations=[40] * pq.ns,
            labels=np.array(["test epoch 2"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations={"index": [1], "test": ["b"]},
            **params,
        )
        targ.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(targ)

        t_start = 0.0012 * pq.s
        t_stop = 0.0016 * pq.s
        result = epc.time_slice(t_start, t_stop)

        assert_arrays_equal(result.times, targ.times)
        assert_arrays_equal(result.durations, targ.durations)
        assert_arrays_equal(result.labels, targ.labels)
        self.assertEqual(result.name, targ.name)
        self.assertEqual(result.description, targ.description)
        self.assertEqual(result.file_origin, targ.file_origin)
        self.assertEqual(result.annotations["test0"], targ.annotations["test0"])
        self.assertEqual(result.annotations["test1"], targ.annotations["test1"])
        self.assertEqual(result.annotations["test2"], targ.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.array([1]))
        assert_arrays_equal(result.array_annotations["test"], np.array(["b"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test__time_slice_should_set_parents_to_None(self):
        # When timeslicing, a deep copy is made,
        # thus the reference to parent objects should be destroyed
        result = self.epc.time_slice(1 * pq.ms, 3 * pq.ms)
        self.assertEqual(result.segment, None)

    def test__deepcopy_should_set_parents_objects_to_None(self):
        # Deepcopy should destroy references to parents
        result = deepcopy(self.epc)
        self.assertEqual(result.segment, None)

    def test__time_shift_same_attributes(self):
        result = self.epc.time_shift(1 * pq.ms)
        assert_same_attributes(result, self.epc, exclude=["times"])

    def test__time_shift_same_annotations(self):
        result = self.epc.time_shift(1 * pq.ms)
        assert_same_annotations(result, self.epc)

    def test__time_shift_same_array_annotations(self):
        result = self.epc.time_shift(1 * pq.ms)
        assert_same_array_annotations(result, self.epc)

    def test__time_shift_should_set_parents_to_None(self):
        # When time-shifting, a deep copy is made,
        # thus the reference to parent objects should be destroyed
        result = self.epc.time_shift(1 * pq.ms)
        self.assertEqual(result.segment, None)

    def test__time_shift_by_zero(self):
        shifted = self.epc.time_shift(0 * pq.ms)
        assert_arrays_equal(shifted.times, self.epc.times)

    def test__time_shift_same_units(self):
        shifted = self.epc.time_shift(10 * pq.ms)
        assert_arrays_equal(shifted.times, self.epc.times + 10 * pq.ms)

    def test__time_shift_different_units(self):
        shifted = self.epc.time_shift(1 * pq.s)
        assert_arrays_equal(shifted.times, self.epc.times + 1000 * pq.ms)

    def test_as_array(self):
        times = [2, 3, 4, 5]
        durations = [0.1, 0.2, 0.3, 0.4]
        epc = Epoch(times * pq.ms, durations=durations * pq.ms)
        epc_as_arr = epc.as_array(units="ms")
        self.assertIsInstance(epc_as_arr, np.ndarray)
        assert_array_equal(times, epc_as_arr)

    def test_as_quantity(self):
        times = [2, 3, 4, 5]
        durations = [0.1, 0.2, 0.3, 0.4]
        epc = Epoch(times * pq.ms, durations=durations * pq.ms)
        epc_as_q = epc.as_quantity()
        self.assertIsInstance(epc_as_q, pq.Quantity)
        assert_array_equal(times * pq.ms, epc_as_q)

    def test_getitem_scalar(self):
        times = [2, 3, 4, 5]
        durations = [0.1, 0.2, 0.3, 0.4]
        labels = ["A", "B", "C", "D"]
        epc = Epoch(times * pq.ms, durations=durations * pq.ms, labels=labels)
        single_epoch = epc[2]
        self.assertIsInstance(single_epoch, pq.Quantity)
        assert_array_equal(single_epoch.times, np.array([4.0]))
        assert_array_equal(single_epoch.durations, np.array([0.3]))
        assert_array_equal(single_epoch.labels, np.array(["C"]))

    def test_slice(self):
        times = [2, 3, 4, 5]
        durations = [0.1, 0.2, 0.3, 0.4]
        labels = ["A", "B", "C", "D"]
        arr_ann = {"index": np.arange(4), "test": ["a", "b", "c", "d"]}
        epc = Epoch(times * pq.ms, durations=durations * pq.ms, labels=labels, array_annotations=arr_ann)
        single_epoch = epc[1:3]
        self.assertIsInstance(single_epoch, Epoch)
        assert_array_equal(single_epoch.times, np.array([3.0, 4.0]))
        assert_array_equal(single_epoch.durations, np.array([0.2, 0.3]))
        assert_array_equal(single_epoch.labels, np.array(["B", "C"]))
        assert_arrays_equal(single_epoch.array_annotations["index"], np.arange(1, 3))
        assert_arrays_equal(single_epoch.array_annotations["test"], np.array(["b", "c"]))
        self.assertIsInstance(single_epoch.array_annotations, ArrayDict)

    def test_rescale(self):
        times = [2, 3, 4, 5]
        durations = [0.1, 0.2, 0.3, 0.4]
        labels = ["A", "B", "C", "D"]
        arr_ann = {"index": np.arange(4), "test": ["a", "b", "c", "d"]}
        epc = Epoch(times * pq.ms, durations=durations * pq.ms, labels=labels, array_annotations=arr_ann)
        result = epc.rescale(pq.us)

        self.assertIsInstance(result, Epoch)
        assert_neo_object_is_compliant(result)
        assert_arrays_equal(result.array_annotations["index"], np.arange(4))
        assert_arrays_equal(result.array_annotations["test"], np.array(["a", "b", "c", "d"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(result.units, 1 * pq.us)
        assert_array_equal(epc.labels, result.labels)
        assert_arrays_almost_equal(result.times, [2000, 3000, 4000, 5000] * pq.us, 1e-9)
        assert_arrays_almost_equal(result.times.magnitude, np.array([2000, 3000, 4000, 5000]), 1e-9)
        assert_arrays_almost_equal(result.durations.magnitude, np.array([100, 200, 300, 400]), 1e-9)


class TestDuplicateWithNewData(unittest.TestCase):
    def setUp(self):
        self.data = np.array([0.1, 0.5, 1.2, 3.3, 6.4, 7])
        self.durations = np.array([0.2, 0.4, 1.1, 2.4, 0.2, 2.0])
        self.quant = pq.ms
        self.arr_ann = {"index": np.arange(6), "test": ["a", "b", "c", "d", "e", "f"]}
        self.epoch = Epoch(
            self.data * self.quant, durations=self.durations * self.quant, array_annotations=self.arr_ann
        )

    def test_duplicate_with_new_data(self):
        signal1 = self.epoch
        new_times = np.sort(np.random.uniform(0, 100, self.epoch.size)) * pq.ms
        new_durations = np.ones_like(new_times)
        new_labels = np.array(list("zyxwvutsrqponmlkjihgfedcba"[: self.epoch.size]))
        signal1b = signal1.duplicate_with_new_data(new_times, new_durations, new_labels)
        # After duplicating, array annotations should always be empty,
        # because different length of data would cause inconsistencies
        assert_arrays_equal(signal1b.labels, new_labels)
        assert_arrays_equal(signal1b.durations, new_durations)
        self.assertTrue("index" not in signal1b.array_annotations)
        self.assertTrue("test" not in signal1b.array_annotations)
        self.assertIsInstance(signal1b.array_annotations, ArrayDict)


class TestEpochFunctions(unittest.TestCase):
    def test__pickle(self):
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        epoch1 = Epoch(
            np.arange(0, 30, 10) * pq.s,
            durations=[1, 2, 3] * pq.s,
            labels=np.array(["t0", "t1", "t2"], dtype="U"),
            units="s",
            annotation1="foo",
            annotation2="bar",
            array_annotations=arr_ann,
        )
        fobj = open("./pickle", "wb")
        pickle.dump(epoch1, fobj)
        fobj.close()

        fobj = open("./pickle", "rb")
        try:
            epoch2 = pickle.load(fobj)
        except ValueError:
            epoch2 = None

        fobj.close()
        assert_array_equal(epoch1.times, epoch2.times)
        self.assertEqual(epoch2.annotations, epoch1.annotations)
        assert_arrays_equal(epoch2.array_annotations["index"], np.array(arr_ann["index"]))
        assert_arrays_equal(epoch2.array_annotations["test"], np.array(arr_ann["test"]))
        self.assertIsInstance(epoch2.array_annotations, ArrayDict)
        # Make sure the dict can perform correct checks after unpickling
        epoch2.array_annotations["anno3"] = list(range(3, 6))
        with self.assertRaises(ValueError):
            epoch2.array_annotations["anno4"] = [2, 1]
        os.remove("./pickle")


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.neuroscopeio
"""

import unittest

from neo.io import NeuroScopeIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestNeuroScopeIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = NeuroScopeIO
    entities_to_download = ["neuroscope"]
    entities_to_test = ["neuroscope/test1/test1.xml"]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.kwikio
"""

import unittest

try:
    from klusta import kwik

    HAVE_KWIK = True
except ImportError:
    HAVE_KWIK = False

from neo.io import kwikio
from neo.test.iotest.common_io_test import BaseTestIO


@unittest.skipUnless(HAVE_KWIK, "requires klusta")
class TestKwikIO(BaseTestIO, unittest.TestCase):
    ioclass = kwikio.KwikIO
    entities_to_download = ["kwik"]
    entities_to_test = ["kwik/neo.kwik"]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.plexonrawio import PlexonRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestPlexonRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = PlexonRawIO
    entities_to_download = ["plexon"]
    entities_to_test = [
        "plexon/File_plexon_1.plx",
        "plexon/File_plexon_2.plx",
        "plexon/File_plexon_3.plx",
    ]


if __name__ == "__main__":
    unittest.main()



"""

"""

import unittest

from neo.rawio.brainvisionrawio import BrainVisionRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestBrainVisionRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = BrainVisionRawIO
    entities_to_test = [
        "brainvision/File_brainvision_1.vhdr",
        "brainvision/File_brainvision_2.vhdr",
        "brainvision/File_brainvision_3_float32.vhdr",
        "brainvision/File_brainvision_3_int16.vhdr",
        "brainvision/File_brainvision_3_int32.vhdr",
    ]

    entities_to_download = ["brainvision"]


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.alphaomegarawio import AlphaOmegaRawIO


class AlphaOmegaIO(AlphaOmegaRawIO, BaseFromRaw):
    """Class for reading data from AlphaOmega MPX file"""

    __doc__ = AlphaOmegaRawIO.__doc__

    def __init__(self, filename, lsx_files=None, prune_channels=True):
        AlphaOmegaRawIO.__init__(
            self,
            dirname=filename,
            lsx_files=lsx_files,
            prune_channels=prune_channels,
        )
        BaseFromRaw.__init__(self, filename)
